import csv
from decimal import Decimal
from io import StringIO

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from chellow.utils import (
    HH,
    ct_datetime_parse,
    loads,
    parse_mpan_core,
    to_utc,
    validate_hh_start,
)


def parse_date(row, idx, is_finish=False):
    date_str = row[idx].strip()
    try:
        if len(date_str) == 10:
            dt = to_utc(ct_datetime_parse(date_str, "%Y-%m-%d"))
            if is_finish:
                dt = dt + relativedelta(days=1) - HH
        else:
            dt = to_utc(ct_datetime_parse(date_str, "%Y-%m-%d %H:%M"))
        return validate_hh_start(dt)
    except ValueError as e:
        raise BadRequest(
            f"Difficulty parsing a date in the row {row} at position " f"{idx}: {e}"
        )


class Parser:
    def __init__(self, f):
        self.reader = csv.reader(
            StringIO(str(f.read(), "utf-8", errors="ignore")), skipinitialspace=True
        )
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        next(iter(self.reader))  # skip title row
        blank_set = set(("",))
        for self.line_number, self.vals in enumerate(self.reader):

            # skip blank lines
            if len(self.vals) == 0 or set(self.vals) == blank_set:
                continue

            bill_type_code = self.vals[0]
            if bill_type_code.startswith("#"):
                continue  # skip comment lines
            account = self.vals[1]
            mpan_core = parse_mpan_core(self.vals[2])
            reference = self.vals[3]
            issue_date = parse_date(self.vals, 4)
            start_date = parse_date(self.vals, 5)
            finish_date = parse_date(self.vals, 6, True)

            kwh = self.to_decimal(7, "kwh")
            net = self.to_decimal(8, "net", True)
            vat = self.to_decimal(9, "vat", True)
            gross = self.to_decimal(10, "gross", True)

            if len(self.vals) > 11:
                breakdown_str = self.vals[11].strip()
                if len(breakdown_str) == 0:
                    breakdown = {}
                else:
                    try:
                        breakdown = loads(breakdown_str)
                    except SyntaxError as e:
                        raise BadRequest(str(e))
            else:
                raise BadRequest(
                    f"For the line, {self.vals} there isn't a 'breakdown' "
                    f"field on the end."
                )

            while self.vals[-1] == "" and len(self.vals) > 12:
                del self.vals[-1]

            reads = []
            for i in range(12, len(self.vals), 11):
                tpr_str = self.vals[i + 4].strip()
                tpr_code = None if len(tpr_str) == 0 else tpr_str.zfill(5)
                reads.append(
                    {
                        "msn": self.vals[i],
                        "mpan": self.vals[i + 1],
                        "coefficient": self.to_decimal(i + 2, "coefficient"),
                        "units": self.vals[i + 3],
                        "tpr_code": tpr_code,
                        "prev_date": parse_date(self.vals, i + 5),
                        "prev_value": Decimal(self.vals[i + 6]),
                        "prev_type_code": self.vals[i + 7],
                        "pres_date": parse_date(self.vals, i + 8),
                        "pres_value": Decimal(self.vals[i + 9]),
                        "pres_type_code": self.vals[i + 10],
                    }
                )

            raw_bill = {
                "bill_type_code": bill_type_code,
                "account": account,
                "mpan_core": mpan_core,
                "reference": reference,
                "issue_date": issue_date,
                "start_date": start_date,
                "finish_date": finish_date,
                "kwh": kwh,
                "net": net,
                "vat": vat,
                "gross": gross,
                "breakdown": breakdown,
                "reads": reads,
            }
            raw_bills.append(raw_bill)

        return raw_bills

    def to_decimal(self, dec_index, dec_name, is_money=False):
        try:
            dec_str = self.vals[dec_index]
            dec = Decimal(dec_str)
            if is_money:
                dec += Decimal("0.00")
            return dec
        except IndexError:
            raise BadRequest(
                f"The field '{dec_name}' can't be found. It's expected at "
                f"position {dec_index} in the list of fields."
            )
        except ValueError:
            raise BadRequest(
                f"The {dec_name} field '{dec_str}' cannot be parsed as a "
                f"number. The {dec_name} field is the {dec_index} field of "
                f"{self.vals}."
            )



import csv
import os
import sys
import threading
import traceback

from flask import g, request

from sqlalchemy import null, or_, true
from sqlalchemy.orm import joinedload

import chellow.dloads
from chellow.models import Era, Session, Site, SiteEra, Source, Supply
from chellow.utils import req_date, req_int, write_row
from chellow.views import chellow_redirect


TYPE_ORDER = {"hh": 0, "amr": 1, "nhh": 2, "unmetered": 3}


def content(sess, start_date, finish_date, site_id, user):
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "site_hh_data.csv", user
        )
        f = open(running_name, mode="w", newline="")
        writer = csv.writer(f, lineterminator="\n")
        write_row(
            writer,
            "Site Id",
            "Site Name",
            "Associated Site Ids",
            "Sources",
            "Generator Types",
            "HH Start Clock-Time",
            "Imported kWh",
            "Displaced kWh",
            "Exported kWh",
            "Used kWh",
            "Parasitic kWh",
            "Generated kWh",
            "3rd Party Import",
            "3rd Party Export",
            "Meter Type",
        )

        sites = sess.query(Site).order_by(Site.code)
        if site_id is not None:
            sites = sites.filter(Site.id == site_id)

        for site in sites:
            sources = set()
            generator_types = set()
            assoc = site.find_linked_sites(sess, start_date, finish_date)
            metering_type = ""
            for era in (
                sess.query(Era)
                .join(SiteEra)
                .filter(
                    SiteEra.site == site,
                    SiteEra.is_physical == true(),
                    Era.start_date <= finish_date,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                    Source.code != "sub",
                )
                .options(
                    joinedload(Era.supply).joinedload(Supply.source),
                    joinedload(Era.supply).joinedload(Supply.generator_type),
                )
            ):
                mtype = era.meter_category
                if metering_type == "" or TYPE_ORDER[mtype] < TYPE_ORDER[metering_type]:
                    metering_type = mtype
                sources.add(era.supply.source.code)
                generator_type = era.supply.generator_type
                if generator_type is not None:
                    generator_types.add(generator_type.code)

            assoc_str = ",".join(sorted(s.code for s in assoc))
            sources_str = ",".join(sorted(list(sources)))
            generators_str = ",".join(sorted(list(generator_types)))

            for hh in site.hh_data(sess, start_date, finish_date):
                write_row(
                    writer,
                    site.code,
                    site.name,
                    assoc_str,
                    sources_str,
                    generators_str,
                    hh["start_date"],
                    hh["imp_net"],
                    hh["displaced"],
                    hh["exp_net"],
                    hh["used"],
                    hh["exp_gen"],
                    hh["imp_gen"],
                    hh["imp_3p"],
                    hh["exp_3p"],
                    metering_type,
                )

    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_date = req_date("start")
    finish_date = req_date("finish")
    if "site_id" in request.values:
        site_id = req_int("site_id")
    else:
        site_id = None

    args = (sess, start_date, finish_date, site_id, g.user)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import datetime
from decimal import Decimal
from io import StringIO

from chellow.utils import parse_mpan_core, validate_hh_start


def parse_date(date_string):
    return validate_hh_start(datetime.datetime.strptime(date_string, "%Y%m%d"))


class Parser:
    def __init__(self, f):
        self.f = StringIO(str(f.read(), "utf-8", errors="ignore"))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        for i, line in enumerate(self.f):
            self.line_number = i
            record_type = line[62:66]
            if record_type == "0100":
                account = line[33:41]
                reference = line[41:46]
                start_date = None
                finish_date = None
                net = Decimal(0)
                vat = Decimal(0)
                mpan_core = None
            elif record_type == "1460":
                net += Decimal(line[67:79]) / Decimal(100)
                vat += Decimal(line[85:97]) / Decimal(100)
            elif record_type == "0461":
                mpan_core = parse_mpan_core(line[135:148])
            elif record_type == "0101":
                start_date = parse_date(line[66:74])
                finish_date = parse_date(line[74:82])
            elif record_type == "1500":
                raw_bill = {
                    "bill_type_code": "N",
                    "mpan_core": mpan_core,
                    "account": account,
                    "reference": reference,
                    "issue_date": start_date,
                    "start_date": start_date,
                    "finish_date": finish_date,
                    "kwh": Decimal("0.00"),
                    "net": net,
                    "vat": vat,
                    "gross": Decimal("0.00"),
                    "breakdown": {},
                    "reads": [],
                }
                raw_bills.append(raw_bill)

        return raw_bills



from dateutil.relativedelta import relativedelta

import chellow.computer
import chellow.duos
from chellow.utils import (
    c_months_u,
    ct_datetime,
    get_file_rates,
    get_file_scripts,
    hh_after,
    hh_before,
    hh_min,
    to_ct,
    to_utc,
)


def hh(ds, rate_period="monthly", est_kw=None):
    for hh in ds.hh_data:
        if hh["ct-is-month-end"]:
            _process_hh(ds, rate_period, est_kw, hh)


def _process_hh(ds, rate_period, est_kw, hh):
    month_start, month_finish = next(
        c_months_u(start_year=hh["ct-year"], start_month=hh["ct-month"])
    )

    month_start_ct = to_ct(month_start)
    if month_start_ct.month > 3:
        year = month_start_ct.year
    else:
        year = month_start_ct.year - 1
    financial_year_start = to_utc(ct_datetime(year, 4, 1))
    last_financial_year_start = to_utc(ct_datetime(year - 1, 4, 1))
    financial_year_finish = to_utc(ct_datetime(year + 1, 3, 31, 23, 30))

    est_triad_kws = []
    earliest_triad = None
    for dt in get_file_rates(ds.caches, "triad_dates", last_financial_year_start)[
        "triad_dates"
    ]:
        triad_hh = None
        earliest_triad = hh_min(earliest_triad, dt)
        try:
            d = next(ds.get_data_sources(dt, dt, financial_year_start))
            chellow.duos.duos_vb(d)
            triad_hh = d.hh_data[0]

            while dt < financial_year_start:
                dt += relativedelta(years=1)

            for d in ds.get_data_sources(dt, dt, financial_year_start):
                chellow.duos.duos_vb(d)
                datum = d.hh_data[0]
                triad_hh["laf"] = datum["laf"]
                triad_hh["gsp-kw"] = datum["laf"] * triad_hh["msp-kw"]
        except StopIteration:
            triad_hh = {
                "hist-start": dt,
                "msp-kw": 0,
                "start-date": dt,
                "status": "before start of MPAN",
                "laf": 1,
                "gsp-kw": 0,
            }
        est_triad_kws.append(triad_hh)

    if ds.site is None:
        era = ds.supply.find_era_at(ds.sess, earliest_triad)
        if (
            era is None
            or era.get_channel(ds.sess, ds.is_import, "ACTIVE") is None
            and est_kw is None
        ):
            est_kw = 0.85 * max(datum["msp-kwh"] for datum in ds.hh_data) * 2
        if est_kw is not None:
            for est_datum in est_triad_kws:
                est_datum["msp-kw"] = est_kw
                est_datum["gsp-kw"] = est_datum["msp-kw"] * est_datum["laf"]

    gsp_kw = 0
    for i, triad_hh in enumerate(est_triad_kws):
        triad_prefix = "triad-estimate-" + str(i + 1)
        hh[triad_prefix + "-date"] = triad_hh["hist-start"]
        hh[triad_prefix + "-msp-kw"] = triad_hh["msp-kw"]
        hh[triad_prefix + "-status"] = triad_hh["status"]
        hh[triad_prefix + "-laf"] = triad_hh["laf"]
        hh[triad_prefix + "-gsp-kw"] = triad_hh["gsp-kw"]
        gsp_kw += triad_hh["gsp-kw"]

    hh["triad-estimate-gsp-kw"] = gsp_kw / 3
    polarity = "import" if ds.llfc.is_import else "export"
    gsp_group_code = ds.gsp_group_code
    rate = float(
        get_file_rates(ds.caches, "triad_rates", month_start)["triad_gbp_per_gsp_kw"][
            polarity
        ][gsp_group_code]
    )

    hh["triad-estimate-rate"] = rate

    est_triad_gbp = hh["triad-estimate-rate"] * hh["triad-estimate-gsp-kw"]

    if rate_period == "monthly":
        total_intervals = 12

        est_intervals = 1
        hh["triad-estimate-months"] = est_intervals
    else:
        dt = financial_year_start
        total_intervals = 0
        while dt <= financial_year_finish:
            total_intervals += 1
            dt += relativedelta(days=1)

        est_intervals = 0
        for d in ds.get_data_sources(month_start, month_finish):
            for h in d.hh_data:
                if h["ct-decimal-hour"] == 0:
                    est_intervals += 1

        hh["triad-estimate-days"] = est_intervals

    hh["triad-estimate-gbp"] = est_triad_gbp / total_intervals * est_intervals

    if hh["ct-month"] == 3:
        triad_kws = []
        for t_date in get_file_rates(ds.caches, "triad_dates", month_start)[
            "triad_dates"
        ]:
            try:
                d = next(ds.get_data_sources(t_date, t_date))
                if (
                    ds.supplier_contract is None
                    or d.supplier_contract == ds.supplier_contract
                ):
                    chellow.duos.duos_vb(d)
                    thh = d.hh_data[0]
                else:
                    thh = {
                        "hist-start": t_date,
                        "msp-kw": 0,
                        "start-date": t_date,
                        "status": "before contract",
                        "laf": "before contract",
                        "gsp-kw": 0,
                    }
            except StopIteration:
                thh = {
                    "hist-start": t_date,
                    "msp-kw": 0,
                    "start-date": t_date,
                    "status": "before start of supply",
                    "laf": "before start of supply",
                    "gsp-kw": 0,
                }

            while t_date < financial_year_start:
                t_date += relativedelta(years=1)

            try:
                d = next(ds.get_data_sources(t_date, t_date))
                if (
                    ds.supplier_contract is None
                    or d.supplier_contract == ds.supplier_contract
                ):
                    chellow.duos.duos_vb(d)
                    thh["laf"] = d.hh_data[0]["laf"]
                    thh["gsp-kw"] = thh["laf"] * thh["msp-kw"]
            except StopIteration:
                pass

            triad_kws.append(thh)
        gsp_kw = 0

        for i, triad_hh in enumerate(triad_kws):
            pref = "triad-actual-" + str(i + 1)
            hh[pref + "-date"] = triad_hh["start-date"]
            hh[pref + "-msp-kw"] = triad_hh["msp-kw"]
            hh[pref + "-status"] = triad_hh["status"]
            hh[pref + "-laf"] = triad_hh["laf"]
            hh[pref + "-gsp-kw"] = triad_hh["gsp-kw"]
            gsp_kw += triad_hh["gsp-kw"]

        hh["triad-actual-gsp-kw"] = gsp_kw / 3
        polarity = "import" if ds.llfc.is_import else "export"
        gsp_group_code = ds.gsp_group_code
        tot_rate = 0
        for start_date, finish_date, script in get_file_scripts("triad_rates"):
            if start_date <= financial_year_finish and not hh_before(
                finish_date, financial_year_start
            ):
                start_month = to_ct(start_date).month
                if start_month < 4:
                    start_month += 12

                if finish_date is None:
                    finish_month = 3
                else:
                    finish_month = to_ct(finish_date).month

                if finish_month < 4:
                    finish_month += 12

                rt = get_file_rates(ds.caches, "triad_rates", start_date)[
                    "triad_gbp_per_gsp_kw"
                ][polarity][gsp_group_code]
                tot_rate += (finish_month - start_month + 1) * float(rt)

        rate = tot_rate / 12
        hh["triad-actual-rate"] = rate

        hh["triad-actual-gbp"] = hh["triad-actual-rate"] * hh["triad-actual-gsp-kw"]

        era = ds.supply.find_era_at(ds.sess, month_finish)
        est_intervals = 0

        interval = (
            relativedelta(months=1)
            if rate_period == "monthly"
            else relativedelta(days=1)
        )

        dt = month_finish
        while era is not None and dt > financial_year_start:
            est_intervals += 1
            dt -= interval
            if hh_after(dt, era.finish_date):
                era = ds.supply.find_era_at(ds.sess, dt)

        if rate_period == "monthly":
            hh["triad-all-estimates-months"] = est_intervals
        else:
            hh["triad-all-estimates-days"] = est_intervals
        hh["triad-all-estimates-gbp"] = (
            est_triad_gbp / total_intervals * est_intervals * -1
        )



from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal
from io import StringIO

from werkzeug.exceptions import BadRequest

from chellow.edi_lib import EdiParser, to_decimal
from chellow.models import Era, Session
from chellow.utils import HH, ct_datetime, parse_mpan_core, to_ct, to_utc


read_type_map = {
    "00": "N",
    "09": "N3",
    "04": "C",
    "02": "E",
    "11": "E3",
    "01": "EM",
    "03": "W",
    "06": "X",
    "05": "CP",
    "12": "IF",
}

tmod_map = {
    "0001": "00001",
    "3U93": "00001",
    "URQ1": "00001",
    "Z012": "00001",
    "246N": "00160",
    "0210": "00210",
    "W246": "00276",
    "0043": "00043",
    "DAQ2": "01071",
    "NIQ2": "01072",
    "NQ22": "00210",
    "DQ22": "00043",
    "D7P2": "00043",
    "N7P2": "00210",
    "MDC1": "kW",
    "URC1": "00001",
    "0206": "00206",
    "0040": "00040",
    "1140": "01140",
    "1139": "01139",
    "A154": "01154",
    "A153": "01153",
    "0039": "00039",
    "0221": "00221",
    "210E": "00210",
    "1150": "01150",
    "1149": "01149",
    "0153": "00153",
    "WQ42": "00184",
    "NQ42": "00210",
    "EQ42": "00187",
    "WDQ3": "01073",
    "EWQ3": "01074",
    "WDQ4": "01075",
    "EWQ4": "01076",
    "NIQ4": "01077",
    "EQ32": "00072",
    "NQ32": "00184",
    "0276": "00276",
    "0277": "00277",
    "0160": "00160",
    "1152": "01152",
    "1151": "01151",
    "1142": "01142",
    "1141": "01141",
    "0249": "00249",
    "0252": "00252",
    "0240": "00240",
    "0244": "00244",
    "0092": "00092",
    "210A": "00210",
    "MAC2": "kW",
    "DAC2": "00040",
    "NAC2": "00206",
    "NDC2": "00040",
    "NNC2": "00206",
    "MNC2": "kW",
    "148A": "00148",
    "221B": "00221",
    "080A": "00080",
    "P1M3": "00248",
    "P2M3": "00251",
    "W1M3": "00239",
    "W2M3": "00073",
    "OTM3": "00093",
    "NIM3": "00208",
    "MDM3": "kW",
    "URQD": "00001",
    "OPE3": "01088",
    "OEC2": "00190",
    "184A": "00184",
    "210J": "00210",
    "187A": "00187",
    "0183": "00183",
    "0071": "00071",
    "DAE7": "00043",
    "NIE7": "00210",
    "OPAC": "00010",
    "MDM1": "kW",
    "D179": "01139",
    "N179": "01140",
    "O212": "00212",
    "0151": "00151",
    "SG1U": "00001",
    "0184": "00184",
    "0072": "00072",
    "DM22": "00039",
    "NM22": "00221",
    "MM22": "kW",
    "0044": "00044",
    "0208": "00208",
    "URM1": "00001",
    "1043": "01043",
    "1042": "01042",
    "U393": "00001",
    "D244": "00040",
    "N244": "00206",
    "24D4": "00040",
    "24N4": "00206",
    "OFX2": "00001",
    "D151": "00043",
    "N151": "00210",
    "N184": "01150",
    "D184": "01149",
    "OPF2": "00257",
    "N154": "00221",
    "O2EC": "00190",
    "OFC2": "00159",
    "CMACUF": "kW",
    "251A": "00251",
    "208D": "00208",
    "073A": "00073",
    "093A": "00093",
    "248A": "00248",
    "239A": "00239",
    "EA1E": "kVA",
    "MDM2": "kVA",
}

WRONG_TPRS = {
    "20 0000 5855 970",
    "20 0000 6048 528",
    "20 0002 1909 377",
    "20 0002 2254 214",
    "20 0002 2326 515",
    "20 0002 2371 155",
    "20 0002 5282 171",
    "20 0002 5476 287",
    "20 0002 6228 157",
    "20 0002 6308 157",
    "20 0002 6419 768",
    "20 0002 6440 184",
    "20 0002 6467 511",
    "22 0001 3834 361",
    "22 0001 4442 321",
    "22 0002 1394 727",
    "22 0002 1401 756",
    "22 0002 1427 578",
    "22 0002 1442 103",
    "22 0002 1502 222",
    "22 0002 1528 986",
    "22 0002 1540 176",
    "22 0002 1589 599",
    "22 0002 1823 142",
    "22 0002 1829 376",
    "22 0002 1839 541",
    "22 0002 1878 160",
}


def to_ct_date(component):
    return to_ct(Datetime.strptime(component, "%y%m%d"))


def to_start_date(component):
    return to_utc(to_ct_date(component))


def to_finish_date(component):
    d = to_ct_date(component)
    return to_utc(ct_datetime(d.year, d.month, d.day, 23, 30))


class Parser:
    def __init__(self, f):
        self.parser = EdiParser(StringIO(str(f.read(), "utf-8", errors="ignore")))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        breakdown = None

        for self.line_number, code in enumerate(self.parser):
            if code == "BCD":
                ivdt = self.parser.elements[0]
                issue_date = to_utc(to_ct_date(ivdt[0]))

                invn = self.parser.elements[2]
                reference = invn[0]
                account = "SA" + reference[:9]

                btcd = self.parser.elements[5]
                bill_type_code = btcd[0]

                sumo = self.parser.elements[7]
                start_date = to_start_date(sumo[0])
                if to_ct_date(sumo[1]) in (
                    ct_datetime(2020, 4, 1),
                    ct_datetime(2020, 3, 16),
                ):
                    finish_date = to_start_date(sumo[1]) - HH
                else:
                    finish_date = to_finish_date(sumo[1])

            elif code == "MHD":
                type = self.parser.elements[1]
                message_type = type[0]
                if message_type == "UTLBIL":
                    issue_date = None
                    start_date = None
                    finish_date = None
                    account = None
                    reference = None
                    net = Decimal("0.00")
                    vat = Decimal("0.00")
                    gross = Decimal("0.00")
                    kwh = Decimal(0)
                    reads = []
                    bill_type_code = None
                    mpan_core = None
                    breakdown = defaultdict(int, {"raw-lines": []})

            elif code == "CCD":
                ccde = self.parser.elements[1]
                consumption_charge_indicator = ccde[0]

                if consumption_charge_indicator == "1":
                    prdt = self.parser.elements[6]
                    pvdt = self.parser.elements[7]

                    pres_read_date = to_finish_date(prdt[0])
                    prev_read_date = to_finish_date(pvdt[0])

                    tmod = self.parser.elements[3]
                    mtnr = self.parser.elements[4]
                    mloc = self.parser.elements[5]

                    mpan = mloc[0]
                    mpan_core = " ".join([mpan[:2], mpan[2:6], mpan[6:10], mpan[10:13]])
                    mpan = (
                        mpan[13:15]
                        + " "
                        + mpan[15:18]
                        + " "
                        + mpan[18:]
                        + " "
                        + mpan_core
                    )

                    prrd = self.parser.elements[9]
                    pres_read_type = read_type_map[prrd[1]]
                    prev_read_type = read_type_map[prrd[3]]

                    adjf = self.parser.elements[12]
                    cons = self.parser.elements[13]

                    coefficient = Decimal(adjf[1]) / Decimal(100000)
                    pres_reading_value = Decimal(prrd[0])
                    prev_reading_value = Decimal(prrd[2])
                    msn = mtnr[0]
                    tpr_native = tmod[0]
                    if tpr_native not in tmod_map:
                        raise BadRequest(
                            "The TPR code "
                            + tpr_native
                            + " can't be found in the TPR list for mpan "
                            + mpan
                            + "."
                        )
                    tpr_code = tmod_map[tpr_native]
                    if tpr_code == "kW":
                        units = "kW"
                        tpr_code = None
                    elif tpr_code == "kVA":
                        units = "kVA"
                        tpr_code = None
                    else:
                        units = "kWh"
                        kwh += to_decimal(cons) / Decimal("1000")

                    if mpan_core in WRONG_TPRS and pres_read_date == to_utc(
                        ct_datetime(2020, 4, 1, 23, 30)
                    ):
                        pres_read_date = to_utc(ct_datetime(2020, 4, 1, 22, 30))
                        reads.append(
                            {
                                "msn": "Separator Read",
                                "mpan": mpan,
                                "coefficient": coefficient,
                                "units": units,
                                "tpr_code": tpr_code,
                                "prev_date": to_utc(ct_datetime(2020, 4, 1, 23)),
                                "prev_value": 0,
                                "prev_type_code": "N",
                                "pres_date": to_utc(ct_datetime(2020, 4, 1, 23)),
                                "pres_value": 0,
                                "pres_type_code": "N",
                            }
                        )
                    reads.append(
                        {
                            "msn": msn,
                            "mpan": mpan,
                            "coefficient": coefficient,
                            "units": units,
                            "tpr_code": tpr_code,
                            "prev_date": prev_read_date,
                            "prev_value": prev_reading_value,
                            "prev_type_code": prev_read_type,
                            "pres_date": pres_read_date,
                            "pres_value": pres_reading_value,
                            "pres_type_code": pres_read_type,
                        }
                    )

                elif consumption_charge_indicator == "2":
                    # tcod = self.parser.elements[2]
                    tmod = self.parser.elements[3]
                    mtnr = self.parser.elements[4]
                    mloc = self.parser.elements[5]

                    mpan = mloc[0]
                    mpan_core = " ".join([mpan[:2], mpan[2:6], mpan[6:10], mpan[10:13]])
                    mpan = (
                        mpan[13:15]
                        + " "
                        + mpan[15:18]
                        + " "
                        + mpan[18:]
                        + " "
                        + mpan_core
                    )

                    prdt = self.parser.elements[6]
                    pvdt = self.parser.elements[7]

                    pres_read_date = to_finish_date(prdt[0])
                    prev_read_date = to_finish_date(pvdt[0])

                    ndrp = self.parser.elements[8]
                    prrd = self.parser.elements[9]
                    pres_read_type = read_type_map[prrd[1]]
                    prev_read_type = read_type_map[prrd[3]]

                    adjf = self.parser.elements[12]
                    cona = self.parser.elements[13]

                    coefficient = Decimal(adjf[1]) / Decimal(100000)
                    pres_reading_value = Decimal(prrd[0])
                    prev_reading_value = Decimal(prrd[2])
                    msn = mtnr[0]
                    tpr_code = tmod[0]
                    if tpr_code not in tmod_map:
                        raise BadRequest(
                            "The TPR code "
                            + tpr_code
                            + " can't be found in the TPR list for mpan "
                            + mpan
                            + "."
                        )
                    tpr = tmod_map[tpr_code]
                    if tpr == "kW":
                        units = "kW"
                        tpr = None
                        prefix = "md-"
                    elif tpr == "kVA":
                        units = "kVA"
                        tpr = None
                        prefix = "md-"
                    else:
                        units = "kWh"
                        kwh += to_decimal(cona) / Decimal("1000")
                        prefix = tpr + "-"

                    nuct = self.parser.elements[15]
                    breakdown[prefix + "kwh"] += to_decimal(nuct) / Decimal("1000")
                    cppu = self.parser.elements[18]
                    rate_key = prefix + "rate"
                    if rate_key not in breakdown:
                        breakdown[rate_key] = set()
                    breakdown[rate_key].add(to_decimal(cppu) / Decimal("100000"))
                    ctot = self.parser.elements[19]
                    breakdown[prefix + "gbp"] += to_decimal(ctot) / Decimal("100")

                    if mpan_core in WRONG_TPRS and pres_read_date == to_utc(
                        ct_datetime(2020, 4, 1, 23, 30)
                    ):
                        pres_read_date = to_utc(ct_datetime(2020, 4, 1, 22, 30))
                        reads.append(
                            {
                                "msn": "Separator Read",
                                "mpan": mpan,
                                "coefficient": coefficient,
                                "units": units,
                                "tpr_code": tpr,
                                "prev_date": to_utc(ct_datetime(2020, 4, 1, 23)),
                                "prev_value": 0,
                                "prev_type_code": "N",
                                "pres_date": to_utc(ct_datetime(2020, 4, 1, 23)),
                                "pres_value": 0,
                                "pres_type_code": "N",
                            }
                        )

                    reads.append(
                        {
                            "msn": msn,
                            "mpan": mpan,
                            "coefficient": coefficient,
                            "units": units,
                            "tpr_code": tpr,
                            "prev_date": prev_read_date,
                            "prev_value": prev_reading_value,
                            "prev_type_code": prev_read_type,
                            "pres_date": pres_read_date,
                            "pres_value": pres_reading_value,
                            "pres_type_code": pres_read_type,
                        }
                    )

                elif consumption_charge_indicator == "3":
                    # tcod = self.parser.elements[2]
                    tmod = self.parser.elements[3]
                    tmod0 = tmod[0]
                    if tmod0 == "CCL":
                        prefix = kwh_prefix = "ccl-"
                    elif tmod0 in ["CQFITC", "CMFITC"]:
                        prefix = "fit-"
                        kwh_prefix = "fit-msp-"
                    elif tmod0 == "FITARR":
                        prefix = kwh_prefix = "fit-reconciliation-"
                    else:
                        tpr_code = tmod0
                        if tpr_code not in tmod_map:
                            raise BadRequest(
                                "The TPR code "
                                + tpr_code
                                + " can't be found in the TPR list for mpan "
                                + mpan
                                + "."
                            )
                        prefix = kwh_prefix = tmod_map[tpr_code] + "-"

                    mtnr = self.parser.elements[4]
                    ndrp = self.parser.elements[8]
                    cona = self.parser.elements[13]
                    nuct = self.parser.elements[15]
                    breakdown[kwh_prefix + "kwh"] += to_decimal(nuct) / Decimal("1000")
                    cppu = self.parser.elements[18]

                    rate_key = prefix + "rate"
                    if rate_key not in breakdown:
                        breakdown[rate_key] = set()
                    breakdown[rate_key].add(to_decimal(cppu) / Decimal("100000"))

                    ctot = self.parser.elements[19]
                    breakdown[prefix + "gbp"] += to_decimal(ctot) / Decimal("100")
                elif consumption_charge_indicator == "4":
                    # tcod = self.parser.elements[2]
                    tmod = self.parser.elements[3]
                    tmod0 = tmod[0]

                    mtnr = self.parser.elements[4]
                    ndrp = self.parser.elements[8]
                    if len(ndrp[0]) > 0:
                        breakdown["standing-days"] += to_decimal(ndrp)
                    cona = self.parser.elements[13]
                    nuct = self.parser.elements[15]
                    cppu = self.parser.elements[18]
                    ctot = self.parser.elements[19]
                    if len(ctot[0]) > 0:
                        breakdown["standing-gbp"] += to_decimal(ctot) / Decimal("100")
            elif code == "MTR":
                if message_type == "UTLBIL":

                    if mpan_core is None:
                        sess = Session()
                        era = (
                            sess.query(Era)
                            .filter(Era.imp_supplier_account == account)
                            .first()
                        )
                        if era is not None:
                            mpan_core = era.imp_mpan_core
                        sess.close()

                    raw_bill = {
                        "bill_type_code": bill_type_code,
                        "account": account,
                        "mpan_core": mpan_core,
                        "reference": reference,
                        "issue_date": issue_date,
                        "start_date": start_date,
                        "finish_date": finish_date,
                        "kwh": kwh,
                        "net": net,
                        "vat": vat,
                        "gross": gross,
                        "breakdown": breakdown,
                        "reads": reads,
                    }
                    raw_bills.append(raw_bill)
                    breakdown = None

            elif code == "MAN":
                madn = self.parser.elements[2]
                """
                pc_code = madn[3]
                mtc_code = madn[4]
                llfc_code = madn[5]
                """

                mpan_core = parse_mpan_core("".join((madn[0], madn[1], madn[2])))

            elif code == "VAT":
                uvla = self.parser.elements[5]
                net += to_decimal(uvla) / Decimal("100")
                uvtt = self.parser.elements[6]
                vat += to_decimal(uvtt) / Decimal("100")
                ucsi = self.parser.elements[7]
                gross += to_decimal(ucsi) / Decimal("100")

            if breakdown is not None:
                breakdown["raw-lines"].append(self.parser.line)

        return raw_bills



import os
import re
import threading
import traceback
from decimal import Decimal, InvalidOperation
from io import BytesIO
from itertools import chain

from flask import g, request

import openpyxl

from werkzeug.exceptions import BadRequest

from zish import dumps

import chellow.dloads
from chellow.models import GspGroup, Session
from chellow.utils import req_int
from chellow.views import chellow_redirect


def get_value(row, idx):
    val = row[idx].value
    if isinstance(val, str):
        return val.strip()
    else:
        return val


def get_rate(row, idx):
    cell = row[idx]
    if cell.number_format == "#":
        return None

    val = cell.value
    if val is None:
        return None
    elif isinstance(val, str) and len(val) == 0:
        return None
    else:
        try:
            return round(Decimal(val) / Decimal("100"), 5)
        except InvalidOperation as e:
            raise BadRequest(f"Can't parse the decimal '{val}' {e}.") from e


def get_rag_rate(row, idx):
    val = get_rate(row, idx)
    return get_rag_rate(row, idx - 1) if val is None else val


def get_zero_rate(row, idx):
    val = get_rate(row, idx)
    return Decimal("0.00000") if val is None else val


def get_decimal(row, idx):
    return round(Decimal(get_value(row, idx)), 5)


def to_llfcs(row, idx):
    val = get_value(row, idx)
    llfcs = []
    if isinstance(val, str):
        for v in map(str.strip, val.split(",")):
            if len(v) > 0:
                if "-" in v:
                    start, finish = v.split("-")
                    m = re.search(r"\d", start)
                    s = m.start()
                    m = re.search(r"\d", finish)
                    f = m.start()

                    for i in range(int(start[s:]), int(finish[f:]) + 1):
                        suff = str(i).zfill(3 - s)
                        llfcs.append(f"{start[:s]}{suff}")
                else:
                    llfcs.append(v)

    elif isinstance(val, Decimal):
        llfcs.append(str(int(val)))
    elif isinstance(val, float):
        llfcs_str = str(int(val))
        for i in range(0, len(llfcs_str), 3):
            llfcs.append(llfcs_str[i : i + 3])
    elif isinstance(val, int):
        val_str = str(val)
        for i in range(0, len(val_str), 3):
            llfcs.append(val_str[i : i + 3])
    return [v.zfill(3) for v in llfcs]


def to_pcs(row, idx):
    val = get_value(row, idx)
    llfcs = []
    if isinstance(val, str):
        for v in map(str.strip, val.replace("or", ",").replace("to", "-").split(",")):
            if len(v) > 0:
                if "-" in v:
                    start, finish = v.split("-")
                    for i in range(int(start), int(finish) + 1):
                        llfcs.append(str(i))
                else:
                    llfcs.append(v)
    elif isinstance(val, Decimal):
        llfcs.append(str(int(val)))
    elif isinstance(val, float):
        llfcs_str = str(int(val))
        for i in range(0, len(llfcs_str), 3):
            llfcs.append(llfcs_str[i : i + 3])
    elif isinstance(val, int):
        val_str = str(val)
        for i in range(0, len(val_str), 3):
            llfcs.append(val_str[i : i + 3])
    return [v.zfill(2) for v in llfcs]


BAND_WEEKEND = {
    "Monday to Friday": False,
    "Weekends": True,
    "Monday to Friday (Including Bank Holidays) All Year": False,
    "Saturday and Sunday All Year": True,
}


def str_to_hr(hr_str):
    for sep in (":", "."):
        if sep in hr_str:
            break
    hours, minutes = map(Decimal, hr_str.strip().split(sep))
    return hours + minutes / Decimal(60)


def val_to_slots(val):
    slots = []
    time_str = None if val is None else str(val).strip()
    if time_str not in (None, "", "[Start] - [End]", "0"):
        for t_str in time_str.splitlines():
            for sep in ("-", "to"):
                if sep in t_str:
                    break
            start_str, finish_str = t_str.split(sep)
            slots.append(
                {"start": str_to_hr(start_str), "finish": str_to_hr(finish_str)}
            )
    return slots


def col_match(row, pattern, repeats=1):
    for i, cell in enumerate(row):
        txt = cell.value
        if txt is not None:
            txt_str = " ".join(str(txt).lower().split())
            if re.search(pattern, txt_str) is not None:
                if repeats == 1:
                    return i
                else:
                    repeats -= 1

    raise BadRequest(
        f"Pattern '{pattern}' not found in row "
        + ", ".join(str(cell.value) for cell in row)
    )


def tab_lv_hv(sheet, gsp_rates):
    try:
        tariffs = gsp_rates["tariffs"]
    except KeyError:
        tariffs = gsp_rates["tariffs"] = {}

    bands = gsp_rates["bands"] = []

    in_tariffs = False
    title_row = None
    for row in sheet.iter_rows():
        val = get_value(row, 0)
        val_0 = None if val is None else " ".join(val.split())
        if in_tariffs:
            if val_0 is None or len(val_0) == 0:
                in_tariffs = False
            else:
                llfcs_str = ",".join(chain(to_llfcs(row, 1), to_llfcs(row, 10)))
                pcs_str = ",".join(to_pcs(row, 2))
                tariffs[llfcs_str + "_" + pcs_str] = {
                    "description": val_0,
                    "gbp-per-mpan-per-day": get_zero_rate(
                        row, col_match(title_row, "fixed")
                    ),
                    "gbp-per-kva-per-day": get_zero_rate(
                        row, col_match(title_row, "^capacity")
                    ),
                    "excess-gbp-per-kva-per-day": get_zero_rate(
                        row, col_match(title_row, "exce")
                    ),
                    "red-gbp-per-kwh": get_rag_rate(row, col_match(title_row, "red")),
                    "amber-gbp-per-kwh": get_rag_rate(
                        row, col_match(title_row, "amber")
                    ),
                    "green-gbp-per-kwh": get_rag_rate(
                        row, col_match(title_row, "green")
                    ),
                    "gbp-per-kvarh": get_zero_rate(
                        row, col_match(title_row, "reactive")
                    ),
                }

        elif val_0 == "Tariff name" or get_value(row, 1) == "Open LLFCs":
            in_tariffs = True
            title_row = row

        if val_0 in BAND_WEEKEND:
            for i, band_name in enumerate(("red", "amber")):
                for slot in val_to_slots(get_value(row, i + 1)):
                    bands.append(
                        {
                            "weekend": BAND_WEEKEND[val_0],
                            "start": slot["start"],
                            "finish": slot["finish"],
                            "band": band_name,
                        }
                    )


# State for EHV

EHV_BLANK = 0
EHV_BANDS = 1
EHV_TARIFFS = 2


def tab_ehv(sheet, gsp_rates):
    try:
        tariffs = gsp_rates["tariffs"]
    except KeyError:
        tariffs = gsp_rates["tariffs"] = {}

    bands = gsp_rates["super_red"] = []

    state = EHV_BLANK
    title_row = None
    for row in sheet.iter_rows():
        val = get_value(row, 0)
        val_0 = None if val is None else " ".join(str(val).split()).lower()
        if state == EHV_BLANK:
            if val_0 == "time periods":
                state = EHV_BANDS
                title_row = row
            elif val_0 in ("import unique identifier", "import llfc"):
                state = EHV_TARIFFS
                title_row = row

        elif state == EHV_TARIFFS:
            for polarity, repeats in (("import", 1), ("export", 2)):
                llfc_val = get_value(row, col_match(title_row, "llfc", repeats=repeats))
                llfc = None if llfc_val is None else str(llfc_val).strip()
                if llfc not in (None, ""):
                    tariffs[llfc] = {
                        "gbp-per-kwh": get_rate(
                            row, col_match(title_row, polarity + " super red")
                        ),
                        "gbp-per-day": get_zero_rate(
                            row, col_match(title_row, polarity + " fixed")
                        ),
                        "gbp-per-kva-per-day": get_zero_rate(
                            row, col_match(title_row, polarity + " capacity")
                        ),
                        "excess-gbp-per-kva-per-day": get_zero_rate(
                            row, col_match(title_row, polarity + " exce")
                        ),
                    }

        elif state == EHV_BANDS:
            if val_0 in (None, "", "notes"):
                state = EHV_BLANK
            else:
                period_str = " ".join(get_value(row, 0).split()).lower()
                periods = []

                if (
                    period_str
                    == "monday to friday nov to feb (excluding "
                    + "22nd dec to 4th jan inclusive)"
                ):
                    periods.append(
                        {
                            "weekend": False,
                            "start-month": 11,
                            "start-day": 1,
                            "finish-month": 12,
                            "finish-day": 21,
                        }
                    )
                    periods.append(
                        {
                            "weekend": False,
                            "start-month": 1,
                            "start-day": 5,
                            "finish-month": 2,
                            "finish-day": "last",
                        }
                    )

                elif period_str in (
                    "monday to friday (including bank "
                    "holidays) november to february",
                    "monday to friday nov to feb",
                ):
                    periods.append(
                        {
                            "weekend": False,
                            "start-month": 11,
                            "start-day": 1,
                            "finish-month": 2,
                            "finish-day": "last",
                        }
                    )

                for slot in val_to_slots(get_value(row, 3)):
                    for period in periods:
                        bands.append(
                            {
                                **period,
                                "start_hour": slot["start"],
                                "finish_hour": slot["finish"],
                            }
                        )


def content(user, file_name, file_like, gsp_group_id):
    f = sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names("dno_rates.zish", user)
        f = open(running_name, mode="w")
        gsp_group = GspGroup.get_by_id(sess, gsp_group_id)
        gsp_rates = {}

        if not file_name.endswith(".xlsx"):
            raise BadRequest(f"The file extension for {file_name} isn't recognized.")

        book = openpyxl.load_workbook(file_like, data_only=True, read_only=True)

        for sheet in book.worksheets:
            title = sheet.title.strip().lower()
            if title.startswith("annex 1 "):
                tab_lv_hv(sheet, gsp_rates)
            elif title.startswith("annex 2 "):
                tab_ehv(sheet, gsp_rates)

        rs = {gsp_group.code: gsp_rates}
        f.write(dumps(rs))

    except BaseException:
        f.write(traceback.format_exc())
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_post(session):
    user = g.user
    file_item = request.files["dno_file"]
    gsp_group_id = req_int("gsp_group_id")

    args = user, file_item.filename, BytesIO(file_item.read()), gsp_group_id
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import os
import threading
import time
import traceback
from datetime import datetime as Datetime

from dateutil.relativedelta import relativedelta

from flask import g, request

import pytz

from sqlalchemy import Float, cast, func, or_
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null, true

import chellow.computer
import chellow.dloads
from chellow.models import Bill, Channel, Era, HhDatum, Session, Site, SiteEra, Supply
from chellow.utils import HH, hh_format, hh_max, hh_min, req_int
from chellow.views import chellow_redirect


def content(year, month, months, supply_id, user):
    tmp_file = sess = None
    try:
        sess = Session()
        supplies = (
            sess.query(Supply)
            .join(Era)
            .distinct()
            .options(joinedload(Supply.generator_type))
        )

        if supply_id is None:
            base_name = (
                "supplies_monthly_duration_for_all_supplies_for_"
                + str(months)
                + "_to_"
                + str(year)
                + "_"
                + str(month)
                + ".csv"
            )
        else:
            supply = Supply.get_by_id(sess, supply_id)
            supplies = supplies.filter(Supply.id == supply.id)
            base_name = (
                "supplies_monthly_duration_for_"
                + str(supply.id)
                + "_"
                + str(months)
                + "_to_"
                + str(year)
                + "_"
                + str(month)
                + ".csv"
            )
        running_name, finished_name = chellow.dloads.make_names(base_name, user)

        tmp_file = open(running_name, "w")

        caches = {}

        start_date = Datetime(year, month, 1, tzinfo=pytz.utc) - relativedelta(
            months=months - 1
        )

        field_names = (
            "supply-name",
            "source-code",
            "generator-type",
            "month",
            "pc-code",
            "msn",
            "site-code",
            "site-name",
            "metering-type",
            "import-mpan-core",
            "metered-import-kwh",
            "metered-import-net-gbp",
            "metered-import-estimated-kwh",
            "billed-import-kwh",
            "billed-import-net-gbp",
            "export-mpan-core",
            "metered-export-kwh",
            "metered-export-estimated-kwh",
            "billed-export-kwh",
            "billed-export-net-gbp",
            "problem",
            "timestamp",
        )

        tmp_file.write("supply-id," + ",".join(field_names) + "\n")

        forecast_date = chellow.computer.forecast_date()

        for i in range(months):
            month_start = start_date + relativedelta(months=i)
            month_finish = month_start + relativedelta(months=1) - HH

            for supply in supplies.filter(
                Era.start_date <= month_finish,
                or_(Era.finish_date == null(), Era.finish_date >= month_start),
            ):

                generator_type = supply.generator_type
                if generator_type is None:
                    generator_type = ""
                else:
                    generator_type = generator_type.code

                source_code = supply.source.code
                eras = supply.find_eras(sess, month_start, month_finish)
                era = eras[-1]
                metering_type = era.meter_category

                site = (
                    sess.query(Site)
                    .join(SiteEra)
                    .filter(SiteEra.era == era, SiteEra.is_physical == true())
                    .one()
                )

                values = {
                    "supply-name": supply.name,
                    "source-code": source_code,
                    "generator-type": generator_type,
                    "month": hh_format(month_finish),
                    "pc-code": era.pc.code,
                    "msn": era.msn,
                    "site-code": site.code,
                    "site-name": site.name,
                    "metering-type": metering_type,
                    "problem": "",
                }

                tmp_file.write(str(supply.id) + ",")

                for is_import, pol_name in [(True, "import"), (False, "export")]:
                    if is_import:
                        mpan_core = era.imp_mpan_core
                    else:
                        mpan_core = era.exp_mpan_core

                    values[pol_name + "-mpan-core"] = mpan_core
                    kwh = 0
                    est_kwh = 0

                    if metering_type in ["hh", "amr"]:
                        est_kwh = (
                            sess.query(HhDatum.value)
                            .join(Channel)
                            .join(Era)
                            .filter(
                                HhDatum.status == "E",
                                Era.supply_id == supply.id,
                                Channel.channel_type == "ACTIVE",
                                Channel.imp_related == is_import,
                                HhDatum.start_date >= month_start,
                                HhDatum.start_date <= month_finish,
                            )
                            .first()
                        )
                        if est_kwh is None:
                            est_kwh = 0
                        else:
                            est_kwh = est_kwh[0]

                    if not (is_import and source_code in ("net", "gen-net")):
                        kwh_sum = (
                            sess.query(cast(func.sum(HhDatum.value), Float))
                            .join(Channel)
                            .join(Era)
                            .filter(
                                Era.supply_id == supply.id,
                                Channel.channel_type == "ACTIVE",
                                Channel.imp_related == is_import,
                                HhDatum.start_date >= month_start,
                                HhDatum.start_date <= month_finish,
                            )
                            .one()[0]
                        )
                        if kwh_sum is not None:
                            kwh += kwh_sum

                    values["metered-" + pol_name + "-estimated-kwh"] = est_kwh
                    values["metered-" + pol_name + "-kwh"] = kwh
                    values["metered-" + pol_name + "-net-gbp"] = 0
                    values["billed-" + pol_name + "-kwh"] = 0
                    values["billed-" + pol_name + "-net-gbp"] = 0
                    values["billed-" + pol_name + "-apportioned-kwh"] = 0
                    values["billed-" + pol_name + "-apportioned-net-gbp"] = 0
                    values["billed-" + pol_name + "-raw-kwh"] = 0
                    values["billed-" + pol_name + "-raw-net-gbp"] = 0

                for bill in sess.query(Bill).filter(
                    Bill.supply == supply,
                    Bill.start_date <= month_finish,
                    Bill.finish_date >= month_start,
                ):
                    bill_start = bill.start_date
                    bill_finish = bill.finish_date
                    bill_duration = (bill_finish - bill_start).total_seconds() + 30 * 60
                    overlap_duration = (
                        min(bill_finish, month_finish) - max(bill_start, month_start)
                    ).total_seconds() + 30 * 60
                    overlap_proportion = float(overlap_duration) / float(bill_duration)
                    values["billed-import-net-gbp"] += overlap_proportion * float(
                        bill.net
                    )
                    values["billed-import-kwh"] += overlap_proportion * float(bill.kwh)

                for era in eras:
                    chunk_start = hh_max(era.start_date, month_start)
                    chunk_finish = hh_min(era.finish_date, month_finish)

                    import_mpan_core = era.imp_mpan_core
                    if import_mpan_core is None:
                        continue

                    supplier_contract = era.imp_supplier_contract

                    if source_code in ["net", "gen-net", "3rd-party"]:
                        supply_source = chellow.computer.SupplySource(
                            sess,
                            chunk_start,
                            chunk_finish,
                            forecast_date,
                            era,
                            True,
                            caches,
                        )

                        values["metered-import-kwh"] += sum(
                            datum["msp-kwh"] for datum in supply_source.hh_data
                        )

                        import_vb_function = supply_source.contract_func(
                            supplier_contract, "virtual_bill"
                        )
                        if import_vb_function is None:
                            values["problem"] += (
                                "Can't find the "
                                "virtual_bill  function in the supplier "
                                "contract. "
                            )
                        else:
                            import_vb_function(supply_source)
                            values[
                                "metered-import-net-gbp"
                            ] += supply_source.supplier_bill["net-gbp"]

                        supply_source.contract_func(era.dc_contract, "virtual_bill")(
                            supply_source
                        )
                        values["metered-import-net-gbp"] += supply_source.dc_bill[
                            "net-gbp"
                        ]

                        mop_func = supply_source.contract_func(
                            era.mop_contract, "virtual_bill"
                        )
                        if mop_func is None:
                            values["problem"] += (
                                " MOP virtual_bill " "function can't be found."
                            )
                        else:
                            mop_func(supply_source)
                            mop_bill = supply_source.mop_bill
                            values["metered-import-net-gbp"] += mop_bill["net-gbp"]
                            if len(mop_bill["problem"]) > 0:
                                values["problem"] += (
                                    " MOP virtual bill problem: " + mop_bill["problem"]
                                )

                values["timestamp"] = int(time.time() * 1000)
                tmp_file.write(
                    ",".join('"' + str(values[name]) + '"' for name in field_names)
                    + "\n"
                )
    except BaseException:
        tmp_file.write(traceback.format_exc())
    finally:
        if sess is not None:
            sess.close()
        tmp_file.close()
        os.rename(running_name, finished_name)


def do_get(sess):
    year = req_int("end_year")
    month = req_int("end_month")
    months = req_int("months")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None
    user = g.user
    threading.Thread(
        target=content, args=(year, month, months, supply_id, user)
    ).start()
    return chellow_redirect("/downloads", 303)



from collections import defaultdict

from werkzeug.exceptions import BadRequest


def make_create_future_func_simple(contract_name, fnames=None):
    def create_future_func_simple(multiplier, constant):
        mult = float(multiplier)
        const = float(constant)
        if fnames is None:

            def future_func(ns):
                new_ns = {}
                for k, v in ns.items():
                    new_ns[k] = float(v) * mult + const

                return new_ns

        else:

            def future_func(ns):
                new_ns = {}
                for fname in fnames:
                    try:
                        new_ns[fname] = ns[fname] * mult + const
                    except KeyError:
                        raise BadRequest(
                            "Can't find "
                            + fname
                            + " in rate script "
                            + str(ns)
                            + " for contract name "
                            + contract_name
                            + " ."
                        )

                return new_ns

        return future_func

    return create_future_func_simple


def make_create_future_func_monthly(contract_name, fnames):
    def create_future_func_monthly(multiplier, constant):
        def future_func(ns):
            new_ns = {}
            for fname in fnames:
                old_result = ns[fname]
                last_value = old_result[sorted(old_result.keys())[-1]]
                new_ns[fname] = defaultdict(
                    lambda: last_value,
                    [(k, v * multiplier + constant) for k, v in old_result.items()],
                )

            return new_ns

        return future_func

    return create_future_func_monthly



import csv
import os
import re
import sys
import threading
import traceback
from decimal import Decimal
from io import BytesIO

from flask import g, request

import openpyxl

from sqlalchemy import null, or_

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.models import Llfc, Party, Session
from chellow.utils import ct_datetime, ct_datetime_now, hh_format, req_int, to_utc
from chellow.views import chellow_redirect


LV_NET = ("LV", False)
LV_SUB = ("LV", True)
HV_NET = ("HV", False)
HV_SUB = ("HV", True)
EHV_NET = ("EHV", False)


VL_LOOKUP = {
    "Low-voltage network": LV_NET,
    "Low Voltage Network": LV_NET,
    "Low-voltage substation": LV_SUB,
    "Low Voltage Substation": LV_SUB,
    "High-voltage network": HV_NET,
    "High Voltage Network": HV_NET,
    "High-voltage substation": HV_SUB,
    "High Voltage Substation": HV_SUB,
    "Greater than 22kV connected - demand": HV_NET,
    "Greater than 22kV connected - generation": HV_NET,
    "33kV generic": EHV_NET,
    "33kV generic (demand)": EHV_NET,
    "33kV generic (generation)": EHV_NET,
    "132/33kV generic": EHV_NET,
    "132kV to 33kV generic": EHV_NET,
    "132/HV connected": HV_NET,
    "132kV generic": EHV_NET,
    "132kV generic (demand)": EHV_NET,
    "132kV generic (generation)": EHV_NET,
    "132kV connected": EHV_NET,
    "EHV connected": EHV_NET,
    "EHV 33kV Export": EHV_NET,
    "EHV 33kV Import": EHV_NET,
    "132/EHV connected": EHV_NET,
}


def get_value(row, idx):
    val = row[idx].value
    return val.strip() if isinstance(val, str) else val


def to_llfcs(val):
    llfcs = []
    if isinstance(val, str):
        val = ",".join(val.splitlines())
        val = val.replace("/", ",").replace("inclusive", ",")
        for v in map(str.strip, val.split(",")):
            if len(v) > 0:
                if "-" in v:
                    start, finish = v.split("-")
                    m = re.search(r"\d", start)
                    s = m.start()
                    m = re.search(r"\d", finish)
                    f = m.start()

                    for i in range(int(start[s:]), int(finish[f:]) + 1):
                        llfcs.append(str(i))
                else:
                    llfcs.append(v)
    elif isinstance(val, Decimal):
        llfcs.append(str(int(val)))
    elif isinstance(val, float):
        llfcs_str = str(int(val))
        for i in range(0, len(llfcs_str), 3):
            llfcs.append(llfcs_str[i : i + 3])
    elif isinstance(val, int):
        val_str = str(val)
        for i in range(0, len(val_str), 3):
            llfcs.append(val_str[i : i + 3])
    return [v.zfill(3) for v in llfcs]


def col_match(row, pattern, repeats=1):
    for i, cell in enumerate(row):
        txt = cell.value
        if txt is not None:
            txt_str = " ".join(str(txt).lower().split())
            if re.search(pattern, txt_str) is not None:
                if repeats == 1:
                    return i
                else:
                    repeats -= 1

    raise BadRequest(
        f"Pattern '{pattern}' not found in row "
        + ", ".join(str(cell.value) for cell in row)
    )


def tab_llfcs(sheet):
    llfcs = []

    in_llfcs = False
    # title_row = None
    for row in sheet.iter_rows():
        val = get_value(row, 0)
        val_0 = None if val is None else " ".join(val.split())
        if in_llfcs:
            if val_0 is None or len(val_0) == 0:
                in_llfcs = False
            else:
                vl, is_substation = VL_LOOKUP[val_0]
                for llfc_code in to_llfcs(get_value(row, 5)):
                    llfc = {
                        "code": llfc_code,
                        "voltage_level": vl,
                        "is_substation": is_substation,
                    }
                    llfcs.append(llfc)

        elif val_0 == "Metered voltage":
            in_llfcs = True
            # title_row = row

    return llfcs


def content(user, file_name, file_like, dno_id):
    f = sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            "voltage_levels_general_importer.csv", user
        )
        f = open(running_name, mode="w", newline="")
        writer = csv.writer(f, lineterminator="\n")
        titles = (
            "# update",
            "llfc",
            "DNO Code",
            "LLFC Code",
            "Valid From",
            "LLFC Description",
            "Voltage Level Code",
            "Is Substation?",
            "Is Import?",
            "Valid To",
        )
        writer.writerow(titles)
        dno = Party.get_by_id(sess, dno_id)

        if not file_name.endswith(".xlsx"):
            raise BadRequest(f"The file extension for {file_name} isn't recognized.")

        book = openpyxl.load_workbook(file_like, data_only=True, read_only=True)

        TITLE_START = "annex 5 "
        ss_llfcs = []
        for sheet in book.worksheets:
            title = sheet.title.strip().lower()
            print(title)
            if title.startswith(TITLE_START):

                # if llfs_sheet is None:
                #     raise BadRequest(
                #         f"Can't find the sheet with LLFCs in. Looking for a "
                #        f"case-insenstive match on sheet titles begining "
                #       with "
                #        f"'{TITLE_START}'.")
                ss_llfcs.extend(tab_llfcs(sheet))

        now_ct = ct_datetime_now()
        if now_ct.month < 4:
            fy_year_ct = now_ct.year - 1
        else:
            fy_year_ct = now_ct.year

        fy_start = to_utc(ct_datetime(fy_year_ct, 4, 1))
        fy_finish = to_utc(ct_datetime(fy_year_ct + 1, 3, 31, 23, 30))
        for ss_llfc in ss_llfcs:
            ss_llfc_code = ss_llfc["code"]
            llfc = (
                sess.query(Llfc)
                .filter(
                    Llfc.dno == dno,
                    Llfc.code == ss_llfc_code,
                    Llfc.valid_from <= fy_finish,
                    or_(Llfc.valid_to == null(), Llfc.valid_to >= fy_start),
                )
                .first()
            )

            if llfc is None:
                raise BadRequest(
                    f"There is no LLFC with the code '{ss_llfc_code}' "
                    f"associated with the DNO {dno.code} from "
                    f"{hh_format(fy_start)} to {hh_format(fy_finish)}."
                )

            row = _make_row(llfc, ss_llfc)
            if row is not None:
                writer.writerow(row)

            # Avoid a long-running transaction
            sess.rollback()

    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def _make_row(llfc, ss_llfc):
    if (
        llfc.voltage_level.code != ss_llfc["voltage_level"]
        or llfc.is_substation != ss_llfc["is_substation"]
    ):

        if llfc.voltage_level.code != ss_llfc["voltage_level"]:
            new_vl = ss_llfc["voltage_level"]
        else:
            new_vl = "{no change}"

        if llfc.is_substation != ss_llfc["is_substation"]:
            new_ss = ss_llfc["is_substation"]
        else:
            new_ss = "{no change}"

        return (
            "update",
            "llfc",
            llfc.dno.dno_code,
            llfc.code,
            hh_format(llfc.valid_from),
            "{no change}",
            new_vl,
            new_ss,
            "{no change}",
            "{no change}",
        )


def do_post(session):
    user = g.user
    file_item = request.files["dno_file"]
    dno_id = req_int("dno_id")

    args = user, file_item.filename, BytesIO(file_item.read()), dno_id
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from chellow.hh_importer import https_handler
from chellow.models import (
    Comm,
    Contract,
    Cop,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import utc_datetime


def test_https_handler(mocker, sess):
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    Mtc.insert(
        sess,
        None,
        "845",
        "HH COP5 And Above With Comms",
        False,
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    dno.insert_llfc(
        sess,
        "521",
        "Export (HV)",
        voltage_level,
        False,
        False,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    era.insert_channel(sess, True, "ACTIVE")

    sess.commit()

    mock_requests = mocker.patch("chellow.hh_importer.requests")
    mock_response = mocker.Mock()
    mock_requests.get.return_value = mock_response
    mock_response.json.return_value = {
        "DataPoints": [{"Flags": 0, "Time": 636188256000000000, "Value": 21}]
    }

    log = []

    def log_f(msg):
        log.append(msg)

    properties = {
        "enabled": True,
        "protocol": "https",
        "download_days": 8,
        "url_template": "https://example.com/?from="
        "{{chunk_start.strftime('%d/%m/%Y')}}&to="
        "{{chunk_finish.strftime('%d/%m/%Y')}}",
        "url_values": {"22 7907 4116 080": {"api_key": "768234ht"}},
    }

    now = utc_datetime(2020, 12, 22)

    https_handler(sess, log_f, properties, dc_contract, now=now)
    expected_log = [
        "Window start: 2020-12-14 00:00",
        "Window finish: 2020-12-21 23:30",
        "Looking at MPAN core 22 7867 6232 781.",
        "Retrieving data from " "https://example.com/?from=14/12/2020&to=21/12/2020.",
        "Finished loading.",
    ]
    assert log == expected_log



from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from chellow.edi_lib import parse_edi, to_date, to_decimal
from chellow.utils import HH, to_ct, to_utc


READ_TYPE_MAP = {"00": "A", "01": "E"}

TCOD_MAP = {
    "Unidentified Gas": {"PPK": "ug"},
    "Commodity": {"PPK": "commodity"},
    "Transportation": {"PPD": "transportation_fixed", "PPK": "transportation_variable"},
    "Gas Flexi": {"PPK": "commodity"},
    "Flex - Gas Flexi (New)": {"PPK": "commodity"},
    "Meter Reading": {"PPD": "meter_read"},
    "Meter Reading Credit Oct 19": {"FIX": "meter_read"},
    "Meter Rental": {"PPD": "metering"},
    "CCL": {"PPK": "ccl"},
    "Consumption Based Administration": {"PPK": "admin_variable"},
    "Swing": {"PPK": "swing"},
}

UNIT_MAP = {"M3": "M3", "HH": "HCUF"}


def _to_finish_date(date_str):
    return to_utc(
        to_ct(Datetime.strptime(date_str, "%y%m%d") + relativedelta(days=1) - HH)
    )


def _process_ADJ(elements, headers):
    adjf = elements["ADJF"]
    if adjf[0] == "CV":
        headers["cv"] = Decimal(adjf[1]) / Decimal(100000)


def _process_BCD(elements, headers):
    ivdt = elements["IVDT"]
    headers["issue_date"] = to_date(ivdt[0])

    invn = elements["INVN"]
    headers["reference"] = invn[0]

    btcd = elements["BTCD"]
    headers["bill_type_code"] = btcd[0]

    sumo = elements["SUMO"]
    headers["start_date"] = to_date(sumo[0])
    headers["finish_date"] = _to_finish_date(sumo[1])


def _process_MHD(elements, headers):
    headers.clear()

    typ = elements["TYPE"]
    message_type = headers["message_type"] = typ[0]
    if message_type == "UTLBIL":
        headers["reads"] = []
        headers["raw_lines"] = []
        headers["breakdown"] = defaultdict(int, {"units_consumed": Decimal(0)})
        headers["kwh"] = Decimal("0.00")
        headers["net"] = Decimal("0.00")
        headers["vat"] = Decimal("0.00")
        headers["gross"] = Decimal("0.00")


def _process_CCD1(elements, headers):
    mtnr = elements["MTNR"]
    msn = mtnr[0]

    mloc = elements["MLOC"]

    # Bug in EDI where MPRN missing in second CCD 1
    if "mprn" not in headers:
        headers["mprn"] = mloc[0]

    prdt = elements["PRDT"]
    pvdt = elements["PVDT"]

    pres_read_date = to_date(prdt[0])
    prev_read_date = to_date(pvdt[0])

    prrd = elements["PRRD"]
    pres_read_value = Decimal(prrd[0])
    pres_read_type = READ_TYPE_MAP[prrd[1]]
    prev_read_value = Decimal(prrd[2])
    prev_read_type = READ_TYPE_MAP[prrd[3]]

    conb = elements["CONB"]
    unit = UNIT_MAP[conb[1]]
    headers["breakdown"]["units_consumed"] += to_decimal(conb) / Decimal("1000")

    adjf = elements["ADJF"]
    correction_factor = Decimal(adjf[1]) / Decimal(100000)

    nuct = elements["NUCT"]

    headers["kwh"] += to_decimal(nuct) / Decimal("1000")

    headers["reads"].append(
        {
            "msn": msn,
            "unit": unit,
            "correction_factor": correction_factor,
            "prev_date": prev_read_date,
            "prev_value": prev_read_value,
            "prev_type_code": prev_read_type,
            "pres_date": pres_read_date,
            "pres_value": pres_read_value,
            "pres_type_code": pres_read_type,
        }
    )


def _process_CCD2(elements, headers):
    breakdown = headers["breakdown"]
    ccde = elements["CCDE"]
    ccde_supplier_code = ccde[2]
    tcod = elements["TCOD"]

    tpref = TCOD_MAP[tcod[1]][ccde_supplier_code]

    bpri = elements["BPRI"]
    rate_key = f"{tpref}_rate"
    if rate_key not in breakdown:
        breakdown[rate_key] = set()
    rate = Decimal(bpri[0]) / Decimal("10000000")
    breakdown[rate_key].add(rate)

    nuct = elements["NUCT"]

    try:
        ctot = elements["CTOT"]
        breakdown[f"{tpref}_gbp"] += to_decimal(ctot) / Decimal("100")

        if ccde_supplier_code == "PPK":
            key = f"{tpref}_kwh"
        elif ccde_supplier_code == "PPD":
            key = f"{tpref}_days"

        breakdown[key] += to_decimal(nuct) / Decimal("1000")
    except KeyError:
        pass


def _process_CCD3(elements, headers):
    breakdown = headers["breakdown"]
    ccde = elements["CCDE"]
    ccde_supplier_code = ccde[2]
    tcod = elements["TCOD"]

    tpref = TCOD_MAP[tcod[1]][ccde_supplier_code]

    bpri = elements["BPRI"]
    rate_key = f"{tpref}_rate"
    if rate_key not in breakdown:
        breakdown[rate_key] = set()
    rate = Decimal(bpri[0]) / Decimal("10000000")
    breakdown[rate_key].add(rate)

    nuct = elements["NUCT"]

    try:
        ctot = elements["CTOT"]
        breakdown[f"{tpref}_gbp"] += to_decimal(ctot) / Decimal("100")

        if ccde_supplier_code == "PPK":
            key = f"{tpref}_kwh"
        elif ccde_supplier_code == "PPD":
            key = f"{tpref}_days"

        breakdown[key] += to_decimal(nuct) / Decimal("1000")
    except KeyError:
        pass


def _process_CCD4(elements, headers):
    breakdown = headers["breakdown"]
    ccde = elements["ccde"]
    ccde_supplier_code = ccde[2]
    tcod = elements["TCOD"]

    tpref = TCOD_MAP[tcod[1]][ccde_supplier_code]

    bpri = elements["BPRI"]
    rate_key = tpref + "_rate"
    if rate_key not in breakdown:
        breakdown[rate_key] = set()
    rate = Decimal(bpri[0]) / Decimal("10000000")
    breakdown[rate_key].add(rate)

    nuct = elements["NUCT"]

    try:
        ctot = elements["CTOT"]
        breakdown[tpref + "_gbp"] += to_decimal(ctot) / Decimal("100")

        if ccde_supplier_code == "PPK":
            key = f"{tpref}_kwh"
        elif ccde_supplier_code == "PPD":
            key = f"{tpref}_days"

        breakdown[key] += to_decimal(nuct) / Decimal("1000")
    except KeyError:
        pass


def _process_MTR(elements, headers):
    if headers["message_type"] == "UTLBIL":
        breakdown = headers["breakdown"]
        for k, v in tuple(breakdown.items()):
            if isinstance(v, set):
                breakdown[k] = sorted(v)

        for read in headers["reads"]:
            read["calorific_value"] = headers["cv"]

        return {
            "raw_lines": "\n".join(headers["raw_lines"]),
            "mprn": headers["mprn"],
            "reference": headers["reference"],
            "account": headers["mprn"],
            "reads": headers["reads"],
            "kwh": headers["kwh"],
            "breakdown": headers["breakdown"],
            "net_gbp": headers["net"],
            "vat_gbp": headers["vat"],
            "gross_gbp": headers["gross"],
            "bill_type_code": headers["bill_type_code"],
            "start_date": headers["start_date"],
            "finish_date": headers["finish_date"],
            "issue_date": headers["issue_date"],
        }


def _process_VAT(elements, headers):
    breakdown = headers["breakdown"]
    vatp = elements["VATP"]
    if "vat_rate" not in breakdown:
        breakdown["vat_rate"] = set()
    breakdown["vat_rate"].add(to_decimal(vatp) / Decimal(100000))

    uvla = elements["UVLA"]
    headers["net"] += to_decimal(uvla) / Decimal("100")
    uvtt = elements["UVTT"]
    headers["vat"] += to_decimal(uvtt) / Decimal("100")
    ucsi = elements["UCSI"]
    headers["gross"] += to_decimal(ucsi) / Decimal("100")


def _process_NOOP(elements, headers):
    pass


CODE_FUNCS = {
    "ADJ": _process_ADJ,
    "BCD": _process_BCD,
    "BTL": _process_NOOP,
    "CCD1": _process_CCD1,
    "CCD2": _process_CCD2,
    "CCD3": _process_CCD3,
    "CCD4": _process_CCD4,
    "CDT": _process_NOOP,
    "CLO": _process_NOOP,
    "END": _process_NOOP,
    "FIL": _process_NOOP,
    "MHD": _process_MHD,
    "MTR": _process_MTR,
    "SDT": _process_NOOP,
    "STX": _process_NOOP,
    "TYP": _process_NOOP,
    "TTL": _process_NOOP,
    "VAT": _process_VAT,
    "VTS": _process_NOOP,
}


class Parser:
    def __init__(self, file_bytes):
        self.edi_str = str(file_bytes, "utf-8", errors="ignore")
        self.line_number = None

    def make_raw_bills(self):
        bills = []
        headers = {}
        bill = None
        for self.line_number, line, seg_name, elements in parse_edi(self.edi_str):

            try:
                func = CODE_FUNCS[seg_name]
            except KeyError:
                raise BadRequest(f"Code {seg_name} not recognized.")

            bill = func(elements, headers)

            if "raw_lines" in headers:
                headers["raw_lines"].append(line)
            if bill is not None:
                bills.append(bill)

        return bills



import os
import time
import traceback
from collections import defaultdict
from collections.abc import Mapping, Set
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from flask import Response, request

from jinja2 import Environment

from pytz import timezone, utc

from werkzeug.exceptions import BadRequest

from zish import ZishException, loads

url_root = ""

HH = relativedelta(minutes=30)
MONTH = relativedelta(months=1)
YEAR = relativedelta(years=1)


def req_str(name):
    try:
        return request.values[name]
    except KeyError:
        raise BadRequest(f"The field {name} is required.")


def req_bool(name):
    try:
        return request.values[name] == "true"
    except KeyError:
        return False


def req_int(name):
    try:
        return int(req_str(name))
    except ValueError as e:
        raise BadRequest(f"Problem parsing the field {name} as an integer: {e}")


def req_zish(name):
    try:
        return loads(req_str(name))
    except ZishException as e:
        raise BadRequest(f"Problem parsing the field {name} as Zish: {e}")


def req_date(prefix, resolution="minute"):
    year = req_int(f"{prefix}_year")
    month = req_int(f"{prefix}_month")
    day = req_int(f"{prefix}_day")

    try:
        if resolution == "day":
            d = ct_datetime(year, month, day)
        elif resolution == "minute":
            hour = req_int(f"{prefix}_hour")
            minute = req_int(f"{prefix}_minute")
            d = ct_datetime(year, month, day, hour, minute)
    except ValueError as e:
        raise BadRequest(f"Problem parsing the date {prefix}: {e}.")

    return to_utc(d)


def req_decimal(name):
    try:
        return Decimal(req_str(name))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the field {name} as a decimal: {e}.")


def req_file(name):
    try:
        return request.files[name]
    except KeyError:
        raise BadRequest(f"The file {name} is required.")


def prev_hh(dt):
    return None if dt is None else dt - HH


def next_hh(dt):
    return None if dt is None else dt + HH


def hh_after(dt1, dt2):
    if dt2 is None:
        return False
    else:
        return True if dt1 is None else dt1 > dt2


def hh_before(dt1, dt2):
    if dt1 is None:
        return False
    else:
        return True if dt2 is None else dt1 < dt2


def get_contract_func(contract, func_name):
    gb = {}
    exec(contract.charge_script, gb)
    return gb.get(func_name)


def req_hh_date(prefix):
    dt = req_date(prefix)
    validate_hh_start(dt)
    return dt


def validate_hh_start(dt):
    if dt.minute not in [0, 30] or dt.second != 0 or dt.microsecond != 0:
        raise BadRequest(
            "The half-hour must start exactly on the hour or half past the hour."
        )
    return dt


def parse_hh_start(start_date_str):
    if len(start_date_str) == 0:
        return None

    try:
        year = int(start_date_str[:4])
        month = int(start_date_str[5:7])
        day = int(start_date_str[8:10])
        hour = int(start_date_str[11:13])
        minute = int(start_date_str[14:16])
        if start_date_str[-1] == "Z":
            dt = utc_datetime(year, month, day, hour, minute)
        else:
            dt = to_utc(ct_datetime(year, month, day, hour, minute))
        return validate_hh_start(dt)
    except ValueError as e:
        raise BadRequest(
            f"Can't parse the date: {start_date_str}. It needs to be of the "
            f"form yyyy-mm-dd hh:MM with an optional Z on the end. {e}"
        )


def parse_mpan_core(mcore):
    mcore = mcore.strip().replace(" ", "")
    if len(mcore) != 13:
        raise BadRequest(f"The MPAN core '{mcore}' must contain exactly 13 digits.")

    for char in mcore:
        if char not in "0123456789":
            raise BadRequest("Each character of an MPAN must be a digit.")

    ps = [3, 5, 7, 13, 17, 19, 23, 29, 31, 37, 41, 43]
    cd = sum(p * int(d) for p, d in zip(ps, mcore[:-1])) % 11 % 10
    if cd != int(mcore[-1]):
        raise BadRequest(
            f"The MPAN core {mcore} is not valid. It fails the checksum test."
        )

    return " ".join([mcore[:2], mcore[2:6], mcore[6:10], mcore[10:]])


def parse_bool(bool_str):
    val = bool_str.strip().lower()
    if val == "true":
        return True
    elif val == "false":
        return False
    else:
        raise BadRequest(f"A boolean must be 'true' or 'false', but got '{bool_str}'.")


def hh_format(dt, ongoing_str="ongoing", with_hh=False):
    if dt is None:
        return (ongoing_str, ongoing_str) if with_hh else ongoing_str
    else:
        if dt.tzinfo is None:
            ts = to_utc(dt)
        else:
            ts = dt
        d = to_ct(ts)
        dt_str = d.strftime("%Y-%m-%d %H:%M")
        if with_hh:
            dc = ct_datetime(d.year, d.month, d.day)
            du = to_utc(dc)
            hh = 0
            while dc.day == d.day and du <= ts:
                hh += 1
                du += HH
                dc = to_ct(du)
            return dt_str, hh
        else:
            return dt_str


CHANNEL_TYPES = "ACTIVE", "REACTIVE_IMP", "REACTIVE_EXP"


def parse_channel_type(channel_type):
    tp = channel_type.upper()
    if tp not in CHANNEL_TYPES:
        raise BadRequest(
            f"The given channel type is '{channel_type}' but it should be one of "
            f"{CHANNEL_TYPES}."
        )
    return tp


def parse_pc_code(code):
    return str(int(code)).zfill(2)


def send_response(
    content,
    args=None,
    status=200,
    content_type="text/csv; charset=utf-8",
    file_name=None,
):
    headers = {}
    if args is None:
        args = ()

    if file_name is not None:
        headers["Content-Disposition"] = f'attachment; filename="{file_name}"'

    return Response(
        content(*args), status=status, content_type=content_type, headers=headers
    )


FORMATS = {
    "year": "%Y",
    "month": "%m",
    "day": "%d",
    "hour": "%H",
    "minute": "%M",
    "full": "%Y-%m-%d %H:%M",
    "date": "%Y-%m-%d",
}

prefix = """
{%- macro input_date(prefix, initial=None, resolution='minute') -%}
  {% if prefix != None %}
    {% set year_field = prefix + '_year' %}
    {% set month_field = prefix + '_month' %}
    {% set day_field = prefix + '_day' %}
    {% set hour_field = prefix + '_hour' %}
    {% set minute_field = prefix + '_minute' %}
  {% else %}
    {% set year_field = 'year' %}
    {% set month_field = 'month' %}
    {% set day_field = 'day' %}
    {% set hour_field = 'hour' %}
    {% set minute_field = 'minute' %}
  {% endif %}

  {% set initial = initial|now_if_none %}

  <input name="{{ year_field }}" maxlength="4" size="4" value="
    {%- if request.values.year_field -%}
      {{ request.values.year_field }}
    {%- else -%}
      {{ initial|hh_format('year') }}
    {%- endif %}">

  {%- if resolution in ['month', 'day', 'hour', 'minute'] -%}
    -<select name="{{ month_field }}">
    {% for month in range(1, 13) -%}
      <option value="{{ "%02i"|format(month) }}"
        {%- if request.values.month_field -%}
          {%- if request.values.month_field|int == month %} selected
          {%- endif -%}
        {%- else -%}
          {%- if initial.month == month %} selected{%- endif -%}
        {%- endif -%}>{{ "%02i"|format(month) }}</option>
    {% endfor %}
    </select>
  {%- endif -%}

  {% if resolution in ['day', 'hour', 'minute'] -%}
    -<select name="{{ day_field }}">
      {% for day in range(1, 32) -%}
        <option value="{{ day }}"
          {%- if request.values.day_field -%}
            {%- if request.values.day_field|int == day %} selected
            {% endif -%}
          {% else %}
            {%- if initial.day == day %} selected{% endif -%}
          {%- endif %}>{{ "%02i"|format(day) }}</option>
      {% endfor -%}
    </select>
  {%- endif -%}

  {% if resolution in ['hour', 'minute'] %}
    <select name="{{ hour_field }}">
      {% for hour in range(24) %}
        <option value="{{ hour }}"
          {%- if request.values.hour_field -%}
            {%- if request.values.hour_field|int == hour %} selected
            {%- endif -%}
          {%- else -%}
            {%- if initial.hour == hour %} selected{%- endif -%}
          {%- endif %}>{{ "%02i"|format(hour) }}</option>
      {%- endfor %}
    </select>
  {%- endif -%}

  {% if resolution == 'minute' -%}
    :<select name="{{ minute_field }}">
      {% for minute in range(0, 31, 30) -%}
        <option value="{{ minute }}"
          {%- if request.values.minute_field %}
            {%- if request.values.minute_field|int == minute %} selected
            {%- endif %}
          {%- else %}
            {%- if initial.minute == minute %} selected{% endif %}
            {%- endif %}>{{ "%02i"|format(minute) }}</option>
      {% endfor %}
    </select>
  {%- endif %}
{%- endmacro -%}

{%- macro input_option(name, item_id, desc, initial=None) -%}
    <option value="{{ item_id }}"
        {%- if request.values.name -%}
            {%- if request.values.name == '' ~ item_id %} selected
            {%- endif -%}
        {%- else -%}
            {%- if initial == item_id %} selected{% endif -%}
            {%- endif -%}>{{ desc }}</option>
{%- endmacro -%}

{% macro input_text(
    name, initial=None, size=None, maxlength=None, placeholder=None) %}
    <input name="{{ name }}" value="
        {%- if request.values.name -%}
            {{ request.values.name }}
        {%- elif initial is not none -%}
            {{initial}}
        {%- endif -%}"
        {%- if size %} size="{{ size }}"{% endif %}
        {%- if placeholder %} placeholder="{{ placeholder }}"{% endif %}
        {%- if maxlength %} maxlength="{{ maxlength }}"{% endif %}>
{%- endmacro -%}

{% macro input_textarea(name, initial, rows, cols, placeholder=None) -%}
  <textarea name="{{ name }}" rows="{{ rows }}" cols="{{ cols }}"
    {%- if placeholder %} placeholder="{{ placeholder }}"{% endif %}>
    {%- if request.values.name -%}
      {{ request.values.name }}
    {%- else -%}
      {{ initial }}
    {%- endif -%}
  </textarea>
{%- endmacro -%}

{%- macro input_checkbox(name, initial) %}
    <input type="checkbox" name="{{ name }}" value="true"
                {%- if request.values.name -%}
                    {%- if request.values.name == 'true' %} checked
                    {%- endif -%}
                {%- else -%}
                    {%- if initial == True %} checked{% endif -%}
                    {%- endif -%}>
{%- endmacro -%}
"""


def hh_format_filter(dt, modifier="full"):
    return "Ongoing" if dt is None else dt.strftime(FORMATS[modifier])


def now_if_none(dt):
    return utc_datetime_now() if dt is None else dt


env = Environment(autoescape=True)

env.filters["hh_format"] = hh_format_filter
env.filters["now_if_none"] = now_if_none

template_cache = {}


def render(template, vals, status_code=200, content_type="text/html"):
    if len(template_cache) > 10000:
        template_cache.clear()
    templ_str = prefix + template
    try:
        templ = template_cache[templ_str]
    except KeyError:
        templ = env.from_string(templ_str)
        template_cache[templ_str] = templ

    vals["request"] = request

    headers = {
        "mimetype": content_type,
        "Date": int(round(time.time() * 1000)),
        "Cache-Control": "no-cache",
    }

    try:
        template_str = templ.render(vals)
    except BaseException:
        raise BadRequest(f"Problem rendering template: {traceback.format_exc()}")

    return Response(template_str, status_code, headers)


def hh_range(caches, start_date, finish_date):
    try:
        return caches["utils"]["hh_range"][start_date][finish_date]
    except KeyError:
        try:
            utils_cache = caches["utils"]
        except KeyError:
            utils_cache = caches["utils"] = {}

        try:
            ranges_cache = utils_cache["hh_range"]
        except KeyError:
            ranges_cache = utils_cache["hh_range"] = {}

        try:
            range_start_cache = ranges_cache[start_date]
        except KeyError:
            range_start_cache = ranges_cache[start_date] = {}

        try:
            datetime_cache = utils_cache["datetime"]
        except KeyError:
            datetime_cache = utils_cache["datetime"] = {}

        rg = []
        dt = start_date
        while dt <= finish_date:
            try:
                dt_stored = datetime_cache[dt]
            except KeyError:
                dt_stored = datetime_cache[dt] = dt

            rg.append(dt_stored)
            dt += HH

        range_tuple = range_start_cache[finish_date] = tuple(rg)
        return range_tuple


def hh_min(a_date, b_date):
    if a_date is None:
        return b_date
    if b_date is None:
        return a_date
    return min(a_date, b_date)


def hh_max(a_date, b_date):
    if a_date is None:
        return a_date
    if b_date is None:
        return b_date
    return max(a_date, b_date)


class keydefaultdict(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


ct = timezone("Europe/London")
root_path = None


def ct_datetime(year, month, day=1, hour=0, minute=0):
    return tz_datetime(ct, year, month, day, hour, minute)


def utc_datetime(year, month, day=1, hour=0, minute=0):
    return tz_datetime(utc, year, month, day, hour, minute)


def tz_datetime(tz, year, month, day=1, hour=0, minute=0):
    return tz.normalize(tz.localize(Datetime(year, month, day, hour, minute)))


def to_tz(tz, dt):
    if dt.tzinfo is None:
        return tz.normalize(tz.localize(dt))
    else:
        return tz.normalize(dt.astimezone(tz))


def to_ct(dt):
    return to_tz(ct, dt)


def to_utc(dt):
    return to_tz(utc, dt)


def utc_datetime_now():
    return Datetime.utcnow().replace(tzinfo=utc)


def ct_datetime_now():
    return to_ct(utc_datetime_now())


def utc_datetime_parse(date_str, format_str):
    return Datetime.strptime(date_str, format_str).replace(tzinfo=utc)


def ct_datetime_parse(date_str, format_str):
    return to_ct(Datetime.strptime(date_str, format_str))


def u_months_u(
    start_year=None, start_month=None, finish_year=None, finish_month=None, months=1
):

    if start_year is None:
        start = utc_datetime(finish_year, finish_month) - relativedelta(
            months=months - 1
        )
    else:
        start = utc_datetime(start_year, start_month)

    if finish_year is None:
        if months is None:
            finish = None
        else:
            finish = start + relativedelta(months=months) - HH
    else:
        finish = utc_datetime(finish_year, finish_month) + MONTH - HH

    dt = start
    while hh_before(dt, finish):
        yield dt, dt + MONTH - HH
        dt += MONTH


def c_months_u(
    start_year=None, start_month=None, finish_year=None, finish_month=None, months=1
):

    for c_m_start_c, c_m_finish_c in c_months_c(
        start_year=start_year,
        start_month=start_month,
        finish_year=finish_year,
        finish_month=finish_month,
        months=months,
    ):
        yield to_utc(c_m_start_c), to_utc(c_m_finish_c)


def c_months_c(
    start_year=None, start_month=None, finish_year=None, finish_month=None, months=1
):

    for u_m_start_u, u_m_finish_u in u_months_u(
        start_year=start_year,
        start_month=start_month,
        finish_year=finish_year,
        finish_month=finish_month,
        months=months,
    ):
        c_m_start_c = ct_datetime(u_m_start_u.year, u_m_start_u.month)
        c_m_finish_c = ct_datetime(
            u_m_finish_u.year, u_m_finish_u.month, u_m_finish_u.day, 23, 30
        )

        yield c_m_start_c, c_m_finish_c


def csv_make_val(v):
    if isinstance(v, (Set, list)):
        if len(v) == 1:
            return csv_make_val(next(iter(v)))
        elif 1 < len(v) < 4:
            vals = set(str(csv_make_val(val)) for val in v)
            return " | ".join(sorted(vals))
        else:
            return ""
    elif isinstance(v, Datetime):
        return hh_format(v)
    elif v is None:
        return ""
    else:
        return v


def make_val(v):
    if isinstance(v, (Set, list)):
        if len(v) == 1:
            return make_val(next(iter(v)))
        elif 1 < len(v) < 4:
            vals = set(str(csv_make_val(val)) for val in v)
            return " | ".join(sorted(vals))
        else:
            return None
    elif isinstance(v, Datetime):
        return to_ct(v)
    else:
        return v


def get_file_script_latest(caches, contract_name):
    try:
        return caches["contract_names"][contract_name]["rs_latest"]
    except KeyError:
        try:
            contract_names = caches["contract_names"]
        except KeyError:
            contract_names = caches["contract_names"] = {}

        try:
            cont_cache = contract_names[contract_name]
        except KeyError:
            cont_cache = contract_names[contract_name] = {}

        try:
            cont = cont_cache["rs_latest"]
        except KeyError:
            cont = list(get_file_scripts(contract_name))[-1]
            cont_cache["rs_latest"] = cont
            return cont


def get_file_script(caches, contract_name, date):
    try:
        return caches["contract_names"][contract_name]["rs"][date]
    except KeyError:
        try:
            contract_names = caches["contract_names"]
        except KeyError:
            contract_names = caches["contract_names"] = {}

        try:
            cont_cache = contract_names[contract_name]
        except KeyError:
            cont_cache = contract_names[contract_name] = {}

        try:
            cont = cont_cache["rs"]
        except KeyError:
            cont = cont_cache["rs"] = {}

        scripts = get_file_scripts(contract_name)
        for i, (start_date, finish_date, script_str) in enumerate(scripts):
            if i + 1 == len(scripts):
                sfinish = finish_date
            else:
                sfinish = scripts[i + 1][0] - HH
            if date >= start_date and not hh_after(date, sfinish):
                try:
                    rs = start_date, sfinish, loads(script_str)
                except ZishException as e:
                    raise BadRequest(
                        f"In the rate script {url_root}industry_contracts/"
                        f"{contract_name}/rate_scripts/"
                        f"{start_date.strftime('%Y%m%d%H%M')} there's the problem {e}."
                    )

                begin_date = hh_max(date - YEAR, start_date)
                end_date = hh_min(date + YEAR, sfinish)
                for hh_date in hh_range(caches, begin_date, end_date):
                    cont[hh_date] = rs

        scripts_start = scripts[0][0]
        if date < scripts_start:
            end_date = hh_min(date + YEAR, scripts_start - HH)
            for hh_date in hh_range(caches, date - YEAR, end_date):
                cont[hh_date] = None

        scripts_finish = scripts[-1][1]
        if hh_after(date, scripts_finish):
            start_date = hh_max(date - YEAR, scripts_finish + HH)
            for hh_date in hh_range(caches, start_date, date + YEAR):
                cont[hh_date] = None

        return cont[date]


def get_file_scripts(contract_name):
    rscripts_path = os.path.join(root_path, "rate_scripts", contract_name)

    rscripts = []
    for rscript_fname in sorted(os.listdir(rscripts_path)):
        if not rscript_fname.endswith(".zish"):
            continue
        try:
            start_str, finish_str = rscript_fname.split(".")[0].split("_")
        except ValueError:
            raise Exception(
                f"The rate script {rscript_fname} in the directory {rscripts_path} "
                f" should consist of two dates separated by an underscore."
            )
        start_date = to_utc(Datetime.strptime(start_str, "%Y%m%d%H%M"))
        if finish_str == "ongoing":
            finish_date = None
        else:
            finish_date = to_utc(Datetime.strptime(finish_str, "%Y%m%d%H%M"))
        with open(os.path.join(rscripts_path, rscript_fname), "r") as f:
            script = f.read()
        rscripts.append((start_date, finish_date, script))
    return tuple(rscripts)


class PropDict(Mapping):
    def __init__(self, location, rate_dict, parent_keys=None):
        if not isinstance(rate_dict, Mapping):
            raise Exception(
                f"The rate_dict must be a mapping, but got a {type(rate_dict)} with "
                f"value {rate_dict}"
            )
        self._location = location
        self._parent_keys = [] if parent_keys is None else parent_keys
        storage = {}
        for k, v in rate_dict.items():
            if isinstance(v, Mapping):
                storage[k] = PropDict(location, v, self._parent_keys + [k])
            else:
                storage[k] = v
                if isinstance(v, list):
                    for i, val in enumerate(tuple(v)):
                        if isinstance(val, Mapping):
                            v[i] = PropDict(
                                location,
                                val,
                                self._parent_keys + ["<list item " + str(i) + ">"],
                            )
        self._storage = storage

    def __getitem__(self, key):
        try:
            return self._storage[key]
        except KeyError:
            try:
                return self._storage["*"]
            except KeyError:
                raise KeyError(
                    f"Can't find the the key {self._parent_keys + [key]} or the "
                    f"wildcard {self._parent_keys + ['*']} in {self._location}"
                )

    def __iter__(self):
        return iter(self._storage)

    def __len__(self):
        return len(self._storage)


def get_file_rates(cache, contract_name, dt):
    try:
        return cache["contract_names"][contract_name][dt]
    except KeyError:
        try:
            contract_names = cache["contract_names"]
        except KeyError:
            contract_names = cache["contract_names"] = {}

        try:
            cont = contract_names[contract_name]
        except KeyError:
            cont = contract_names[contract_name] = {}

        try:
            return cont[dt]
        except KeyError:
            rs = get_file_script(cache, contract_name, dt)

            if rs is None:
                l_start, l_finish, l_script = get_file_script_latest(
                    cache, contract_name
                )
                rs = get_file_script(cache, contract_name, l_start)
                hist_date = l_start
            else:
                hist_date = dt
            rs_start, rs_finish, script = rs
            """
            if rs is None:
                dt_weekday = dt.weekday()
                l_rs = get_file_script_latest(cache, contract_name)
                l_start, l_finish, l_script = l_rs
                new_year = l_finish.year
                rs = get_file_script(
                    cache, contract_name, utc_datetime(new_year, dt.month, 7)
                )
                if rs is None:
                    new_year -= 1

                new_dt = utc_datetime(new_year, dt.month, 1, dt.hour, dt.minute)

                for i in range(7):
                    if new_dt.weekday() == dt_weekday:
                        rs = get_file_script(cache, contract_name, new_dt)
                        break
                    new_dt += Timedelta(days=1)

                if rs is None:
                    rs_start, rs_finish, script = l_rs
                    hist_date = rs_start
                else:
                    rs_start, rs_finish, script = rs
                    hist_date = new_dt

            else:
                rs_start, rs_finish, script = rs
                hist_date = dt
            """

            try:
                props = cont["props"]
            except KeyError:
                props = cont["props"] = {}

            try:
                rscript = props[rs_start]
            except KeyError:
                try:
                    script["hist_dates"] = {}
                    rscript = PropDict(
                        f" in the rate script {url_root} industry_contracts/"
                        f"{contract_name}/rate_scripts/"
                        f"{rs_start.strftime('%Y%m%d%H%M')}.",
                        script,
                    )
                except ZishException as e:
                    raise BadRequest(
                        f"Problem parsing rate script for contract {contract_name} "
                        f"starting at {hh_format(rs_start)}: {e}"
                    )
                except BadRequest as e:
                    raise BadRequest(
                        f"Problem with rate script for contract {contract_name} "
                        f"starting at {hh_format(rs_start)}: {e.description}"
                    )

                props[rs_start] = rscript

            rscript["hist_dates"]._storage[dt] = hist_date

            cont[dt] = rscript
            return rscript


def reduce_bill_hhs(bill_hhs):
    bill = {}
    for bill_hh in bill_hhs.values():
        for k, v in bill_hh.items():
            if isinstance(v, set):
                if k in bill:
                    bill[k].update(v)
                else:
                    bill[k] = v

            else:
                if k in bill:
                    bill[k] += v
                else:
                    bill[k] = v

    return bill


def write_row(writer, *vals):
    writer.writerow(csv_make_val(v) for v in vals)



from collections import defaultdict
from itertools import chain

from sqlalchemy import false, true
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

from chellow.models import (
    Batch,
    Bill,
    Channel,
    Contract,
    Era,
    MeasurementRequirement,
    Party,
    RegisterRead,
    Site,
    SiteEra,
    Supply,
    Tpr,
)


def get_era_bundles(sess, supply):
    era_bundles = []
    eras = (
        sess.query(Era)
        .filter(Era.supply == supply)
        .order_by(Era.start_date.desc())
        .options(
            joinedload(Era.pc),
            joinedload(Era.imp_supplier_contract),
            joinedload(Era.exp_supplier_contract),
            joinedload(Era.ssc),
            joinedload(Era.mtc),
            joinedload(Era.mop_contract),
            joinedload(Era.dc_contract),
            joinedload(Era.imp_llfc),
            joinedload(Era.exp_llfc),
            joinedload(Era.supply).joinedload(Supply.dno),
        )
        .all()
    )
    for era in eras:
        imp_mpan_core = era.imp_mpan_core
        exp_mpan_core = era.exp_mpan_core
        physical_site = (
            sess.query(Site)
            .join(SiteEra)
            .filter(SiteEra.is_physical == true(), SiteEra.era == era)
            .one()
        )
        other_sites = (
            sess.query(Site)
            .join(SiteEra)
            .filter(SiteEra.is_physical != true(), SiteEra.era == era)
            .all()
        )
        imp_channels = (
            sess.query(Channel)
            .filter(Channel.era == era, Channel.imp_related == true())
            .order_by(Channel.channel_type)
            .all()
        )
        exp_channels = (
            sess.query(Channel)
            .filter(Channel.era == era, Channel.imp_related == false())
            .order_by(Channel.channel_type)
            .all()
        )
        era_bundle = {
            "era": era,
            "physical_site": physical_site,
            "other_sites": other_sites,
            "imp_channels": imp_channels,
            "exp_channels": exp_channels,
            "imp_bills": {"bill_dicts": []},
            "exp_bills": {"bill_dicts": []},
            "dc_bills": {"bill_dicts": []},
            "mop_bills": {"bill_dicts": []},
        }
        era_bundles.append(era_bundle)

        if imp_mpan_core is not None:
            era_bundle["imp_shared_supplier_accounts"] = (
                sess.query(Supply)
                .distinct()
                .join(Era)
                .filter(
                    Supply.id != supply.id,
                    Era.imp_supplier_account == era.imp_supplier_account,
                    Era.imp_supplier_contract == era.imp_supplier_contract,
                )
                .all()
            )
        if exp_mpan_core is not None:
            era_bundle["exp_shared_supplier_accounts"] = (
                sess.query(Supply)
                .join(Era)
                .filter(
                    Era.supply != supply,
                    Era.exp_supplier_account == era.exp_supplier_account,
                    Era.exp_supplier_contract == era.exp_supplier_contract,
                )
                .all()
            )
        if era.pc.code != "00":
            inner_headers = [
                tpr
                for tpr in sess.query(Tpr)
                .join(MeasurementRequirement)
                .filter(MeasurementRequirement.ssc == era.ssc)
                .order_by(Tpr.code)
            ]
            if era.pc.code in ["05", "06", "07", "08"]:
                inner_headers.append(None)
            era_bundle["imp_bills"]["inner_headers"] = inner_headers
            inner_header_codes = [
                tpr.code if tpr is not None else "md" for tpr in inner_headers
            ]

        bills = (
            sess.query(Bill)
            .filter(Bill.supply == supply)
            .order_by(
                Bill.start_date.desc(), Bill.issue_date.desc(), Bill.reference.desc()
            )
            .options(
                joinedload(Bill.batch)
                .joinedload(Batch.contract)
                .joinedload(Contract.party)
                .joinedload(Party.market_role),
                joinedload(Bill.bill_type),
            )
        )
        if era.finish_date is not None and era != eras[0]:
            bills = bills.filter(Bill.start_date <= era.finish_date)
        if era != eras[-1]:
            bills = bills.filter(Bill.start_date >= era.start_date)

        num_outer_cols = 0
        for bill in bills:
            bill_contract = bill.batch.contract
            bill_role_code = bill_contract.party.market_role.code
            if bill_role_code == "X":
                if (
                    exp_mpan_core is not None
                    and bill_contract == era.exp_supplier_contract
                ):
                    bill_group_name = "exp_bills"
                else:
                    bill_group_name = "imp_bills"

            elif bill_role_code in ("C", "D"):
                bill_group_name = "dc_bills"
            elif bill_role_code == "M":
                bill_group_name = "mop_bills"
            else:
                raise BadRequest(
                    f"Bill group name not found for bill_contract_id "
                    f"{bill_contract.id}."
                )

            bill_group = era_bundle[bill_group_name]
            rows_high = 1
            bill_dict = {"bill": bill}
            bill_group["bill_dicts"].append(bill_dict)

            if bill_group_name == "imp_bills" and era.pc.code != "00":
                inner_tpr_map = dict((code, []) for code in inner_header_codes)
                outer_tpr_map = defaultdict(list)

                for read, tpr in (
                    sess.query(RegisterRead, Tpr)
                    .join(Tpr)
                    .filter(RegisterRead.bill == bill)
                    .order_by(Tpr.id, RegisterRead.present_date.desc())
                    .options(
                        joinedload(RegisterRead.previous_type),
                        joinedload(RegisterRead.present_type),
                        joinedload(RegisterRead.tpr),
                    )
                ):
                    tpr_code = "md" if tpr is None else tpr.code
                    try:
                        inner_tpr_map[tpr_code].append(read)
                    except KeyError:
                        outer_tpr_map[tpr_code].append(read)

                rows_high = max(
                    chain(
                        map(len, chain(inner_tpr_map.values(), outer_tpr_map.values())),
                        [rows_high],
                    )
                )

                read_rows = []
                bill_dict["read_rows"] = read_rows

                for i in range(rows_high):
                    inner_reads = []
                    row_dict = {"inner_reads": inner_reads, "outer_reads": []}
                    read_rows.append(row_dict)
                    for tpr_code in inner_header_codes:
                        try:
                            inner_reads.append(inner_tpr_map[tpr_code][i])
                        except IndexError:
                            row_dict["inner_reads"].append(None)

                    for tpr_code, read_list in outer_tpr_map.items():
                        try:
                            row_dict["outer_reads"].append(read_list[i])
                        except IndexError:
                            row_dict["outer_reads"].append(None)

                num_outer_cols = max(num_outer_cols, len(outer_tpr_map))

                bill_dict["rows_high"] = rows_high

        era_bundle["imp_bills"]["num_outer_cols"] = num_outer_cols
        era_bundle["exp_bills"]["num_outer_cols"] = 0

        for bill_group_name in ("imp_bills", "exp_bills", "dc_bills", "mop_bills"):
            b_dicts = list(reversed(era_bundle[bill_group_name]["bill_dicts"]))
            for i, b_dict in enumerate(b_dicts):
                if i < (len(b_dicts) - 1):
                    bill = b_dict["bill"]
                    next_b_dict = b_dicts[i + 1]
                    next_bill = next_b_dict["bill"]
                    if (
                        (
                            bill.start_date,
                            bill.finish_date,
                            bill.kwh,
                            bill.net,
                            bill.vat,
                        )
                        == (
                            next_bill.start_date,
                            next_bill.finish_date,
                            -1 * next_bill.kwh,
                            -1 * next_bill.net,
                            next_bill.vat,
                        )
                        and not ((bill.kwh, bill.net, bill.vat) == (0, 0, 0))
                        and "collapsible" not in b_dict
                    ):
                        b_dict["collapsible"] = True
                        next_b_dict["first_collapsible"] = True
                        next_b_dict["collapsible"] = True
                        b_dict["collapse_id"] = next_b_dict["collapse_id"] = bill.id
    return era_bundles



import csv
import os
import threading
import traceback
from collections import OrderedDict
from datetime import datetime as Datetime, timedelta as Timedelta
from decimal import Decimal
from io import BytesIO, StringIO
from zipfile import ZipFile

from flask import g, request

from zish import dumps

import chellow.dloads
from chellow.models import Session
from chellow.utils import to_ct, to_utc
from chellow.views import chellow_redirect


def content(user, file_name, file_like):
    f = sess = None
    try:
        sess = Session()
        tps = {}
        llfc_tp = {}
        block = {"llfc_tp": llfc_tp, "tps": tps}
        llfc_code = line_dt = start_date_str = None
        tp_cand = {}
        llfc_data = OrderedDict()
        zip_file = ZipFile(file_like)
        name_list = zip_file.namelist()
        if len(name_list) != 1:
            raise Exception("The zip archive must contain exactly one file.")
        csv_file = StringIO(zip_file.read(name_list[0]).decode("utf-8"))
        for vals in csv.reader(csv_file, delimiter="|"):
            code = vals[0]

            if code in ("LLF", "ZPT"):
                if llfc_code is not None:

                    # Compress days
                    days = OrderedDict()
                    for dt, slots in llfc_data.items():
                        day = days[dt] = []
                        prev_laf = None
                        for slot, laf in slots.items():
                            if laf == prev_laf:
                                day[-1]["slot_finish"] = slot
                            else:
                                day.append(
                                    {
                                        "slot_start": slot,
                                        "slot_finish": slot,
                                        "laf": laf,
                                    }
                                )
                            prev_laf = laf

                    prev_day = last_block = None
                    for dt, day in days.items():
                        if day == prev_day:
                            last_block["finish_date"] = dt
                        else:
                            last_block = tp_cand[dt] = {
                                "start_date": dt,
                                "finish_date": dt,
                                "slots": day,
                            }
                        prev_day = day

                    for tp_id, tp in tps.items():
                        if tp_cand == tp:
                            llfc_tp[llfc_code] = tp_id
                            tp_cand = {}
                            break

                    if tp_cand != {}:
                        tp_id = len(tps)
                        tps[tp_id] = tp_cand
                        llfc_tp[llfc_code] = tp_id

                    tp_cand = {}
                if code == "LLF":
                    llfc_code = vals[1]

            elif code == "SDT":
                line_dt = vals[1]
                if start_date_str is None:
                    start_date_str = line_dt
                llfc_data[line_dt] = OrderedDict()

            elif code == "SPL":
                slot, laf = vals[1:]
                llfc_data[line_dt][slot] = laf

        start_date_raw = Datetime.strptime(start_date_str, "%Y%m%d")
        start_date_ct = to_ct(start_date_raw)
        start_date = to_utc(start_date_ct)

        finish_date_raw = Datetime.strptime(line_dt, "%Y%m%d")
        finish_date_ct = to_ct(finish_date_raw)
        finish_date_ct += Timedelta(minutes=30 * (int(slot) - 1))
        finish_date = to_utc(finish_date_ct)

        running_name, finished_name = chellow.dloads.make_names(
            start_date.strftime("%Y%m%d%H%M")
            + "_"
            + finish_date.strftime("%Y%m%d%H%M")
            + ".zish",
            user,
        )
        f = open(running_name, mode="w")

        llfc_tp = dict((k.zfill(3), v) for k, v in block["llfc_tp"].items())
        block["llfc_tp"] = llfc_tp

        for tp in block["tps"].values():
            for date_block in tp.values():
                for slot in date_block["slots"]:
                    slot["laf"] = Decimal(slot["laf"])
                    slot["slot_start"] = int(slot["slot_start"])
                    slot["slot_finish"] = int(slot["slot_finish"])

        f.write(dumps(block))
    except BaseException:
        f.write(traceback.format_exc())
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_post(session):
    user = g.user
    file_item = request.files["laf_file"]

    args = user, file_item.filename, BytesIO(file_item.read())
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from dateutil.relativedelta import relativedelta

from sqlalchemy import func, select
from sqlalchemy.sql.expression import true

from werkzeug.exceptions import BadRequest

from chellow.models import Channel, Era, HhDatum, Laf, Llfc, Party
from chellow.utils import (
    HH,
    c_months_u,
    ct_datetime,
    get_file_rates,
    hh_format,
    to_utc,
    utc_datetime,
)


BANDS = ("red", "amber", "green")

KEYS = dict(
    (
        band,
        {
            "kwh": f"duos-{band}-kwh",
            "tariff-rate": f"{band}-gbp-per-kwh",
            "bill-rate": f"duos-{band}-rate",
            "gbp": f"duos-{band}-gbp",
        },
    )
    for band in BANDS
)

VL_LOOKUP = {
    "LV": {True: "lv-sub", False: "lv-net"},
    "HV": {True: "hv", False: "hv"},
    "EHV": {True: "ehv", False: "ehv"},
}


def datum_beginning_22(ds, hh):
    rs = get_file_rates(ds.caches, ds.dno_code, hh["start-date"])
    try:
        tariff = rs["tariffs"][ds.llfc_code]
        lafs = rs["lafs"][VL_LOOKUP[ds.voltage_level_code][ds.is_substation]]
    except KeyError as e:
        raise BadRequest(str(e))

    if ds.is_import:
        try:
            day_rate = float(tariff["day-gbp-per-kwh"])
        except KeyError as e:
            raise BadRequest(str(e))

        night_rate = float(tariff["night-gbp-per-kwh"])
        if 6 < hh["ct-decimal-hour"] <= 23:
            hh["duos-day-kwh"] = hh["msp-kwh"]
            hh["duos-day-gbp"] = hh["msp-kwh"] * day_rate
        else:
            hh["duos-night-kwh"] = hh["msp-kwh"]
            hh["duos-night-gbp"] = hh["msp-kwh"] * night_rate

    if 23 < hh["ct-decimal-hour"] <= 6:
        slot_name = "night"
    elif hh["ct-day-of-week"] < 5 and (hh["ct-month"] > 10 or hh["ct-month"] < 3):
        if 15.5 < hh["ct-decimal-hour"] < 18:
            slot_name = "winter-weekday-peak"
        elif 6 < hh["ct-decimal-hour"] < 15:
            slot_name = "winter-weekday-day"
        else:
            slot_name = "other"
    else:
        slot_name = "other"

    try:
        hh["laf"] = float(lafs[slot_name])
    except KeyError as e:
        raise BadRequest(str(e))

    hh["gsp-kwh"] = hh["laf"] * hh["msp-kwh"]
    hh["gsp-kw"] = hh["gsp-kwh"] * 2

    if hh["ct-is-month-end"]:
        month_to = hh["start-date"]
        month_from = month_to - relativedelta(months=1) + HH
        days_in_month = 0
        md_kva = 0
        month_imp_kvarh = 0
        month_kwh = 0
        for dsc in ds.get_data_sources(month_from, month_to):
            for h in dsc.hh_data:
                if h["ct-decimal-hour"] == 0:
                    days_in_month += 1
                md_kva = max(md_kva, (h["msp-kw"] ** 2 + h["imp-msp-kvar"] ** 2) ** 0.5)
                month_imp_kvarh += h["imp-msp-kvarh"]
                month_kwh += h["msp-kwh"]

        tariff = get_file_rates(ds.caches, ds.dno_code, hh["start-date"])["tariffs"][
            ds.llfc_code
        ]
        try:
            reactive_rate = float(tariff["reactive-gbp-per-kvarh"])
        except KeyError as e:
            raise BadRequest(str(e))
        hh["duos-reactive-rate"] = reactive_rate

        if not ds.is_displaced:
            hh["duos-availability-kva"] = ds.sc
            hh["duos-excess-availability-kva"] = max(md_kva - ds.sc, 0)
            for prefix in ["", "excess-"]:
                tariff_key = prefix + "gbp-per-kva-per-day"
                if tariff_key in tariff:
                    rate_key = "duos-" + prefix + "availability-rate"
                    hh[rate_key] = float(tariff[tariff_key])
                    hh["duos-" + prefix + "availability-days"] = days_in_month
                    hh["duos-" + prefix + "availability-gbp"] = (
                        hh[rate_key]
                        * hh["duos-" + prefix + "availability-kva"]
                        * hh["duos-" + prefix + "availability-days"]
                    )

        hh["duos-reactive-gbp"] = (
            max(0, month_imp_kvarh - month_kwh / 2) * reactive_rate
        )


def datum_beginning_20(ds, hh):
    tariff = None
    for k, tf in get_file_rates(ds.caches, ds.dno_code, hh["start-date"])[
        "tariffs"
    ].items():
        if ds.llfc_code in [cd.strip() for cd in k.split(",")]:
            tariff = tf

    if tariff is None:
        raise BadRequest(
            "The tariff for the LLFC "
            + ds.llfc_code
            + " cannot be found for the DNO 20 at "
            + hh_format(hh["start-date"])
            + "."
        )

    lafs = get_file_rates(ds.caches, ds.dno_code, hh["start-date"])["lafs"][
        ds.voltage_level_code.lower()
    ]

    try:
        day_rate = float(tariff["day-gbp-per-kwh"])
    except KeyError as e:
        raise BadRequest(str(e))

    if "night-gbp-per-kwh" in tariff:
        night_rate = float(tariff["night-gbp-per-kwh"])
        if 0 < hh["ct-decimal-hour"] <= 7:
            hh["duos-night-kwh"] = hh["msp-kwh"]
            hh["duos-night-gbp"] = hh["msp-kwh"] * night_rate
        else:
            hh["duos-day-kwh"] = hh["msp-kwh"]
            hh["duos-day-gbp"] = hh["msp-kwh"] * day_rate
    else:
        hh["duos-day-kwh"] = hh["msp-kwh"]
        hh["duos-day-gbp"] = hh["msp-kwh"] * day_rate

    if 0 < hh["ct-decimal-hour"] <= 7:
        slot_name = "night"
    elif (
        hh["ct-day-of-week"] < 5
        and 16 < hh["ct-decimal-hour"] <= 19
        and (hh["ct-month"] > 10 or hh["ct-month"] < 3)
    ):
        slot_name = "peak"
    elif (
        7 > hh["ct-day-of-week"] > 1
        and (7 < hh["ct-decimal-hour"] < 15 or 18.5 < hh["ct-decimal-hour"] < 19)
        and (hh["ct-month"] > 11 or hh["ct-month"] < 4)
    ):
        slot_name = "winter-weekday"
    else:
        slot_name = "other"
    hh["laf"] = float(lafs[slot_name])
    hh["gsp-kwh"] = hh["laf"] * hh["msp-kwh"]
    hh["gsp-kw"] = hh["gsp-kwh"] * 2

    if hh["ct-is-month-end"]:
        tariff = None
        for k, tf in get_file_rates(ds.caches, ds.dno_code, hh["start-date"])[
            "tariffs"
        ].items():
            if ds.llfc_code in map(str.strip, k.split(",")):
                tariff = tf
                break
        if tariff is None:
            raise BadRequest(
                "The tariff for the LLFC "
                + ds.llfc_code
                + " cannot be found for the DNO 20 at "
                + hh_format(hh["start-date"])
                + "."
            )
        if not ds.is_displaced:
            year_md_kva_095 = year_md_095(ds, ds.finish_date)

            hh["duos-excess-availability-kva"] = max(year_md_kva_095 - ds.sc, 0)
            billed_avail = max(ds.sc, year_md_kva_095)
            hh["duos-availability-kva"] = ds.sc

            for threshold, block in [
                (15, 15),
                (100, 5),
                (250, 10),
                (500, 25),
                (1000, 50),
                (None, 100),
            ]:
                if threshold is None or billed_avail < threshold:
                    if billed_avail % block > 0:
                        billed_avail = (int(billed_avail / block) + 1) * block
                    break
            try:
                le_200_avail_rate = float(
                    tariff["capacity-<=200-gbp-per-kva-per-month"]
                )
            except KeyError as e:
                raise BadRequest(str(e))

            hh["duos-availability-gbp"] = min(200, billed_avail) * le_200_avail_rate

            if billed_avail > 200:
                try:
                    gt_200_avail_rate = float(
                        tariff["capacity->200-gbp-per-kva-per-month"]
                    )
                except KeyError as e:
                    raise BadRequest(str(e))

                hh["duos-availability-gbp"] += (billed_avail - 200) * gt_200_avail_rate

        try:
            if "fixed-gbp-per-month" in tariff:
                hh["duos-standing-gbp"] = float(tariff["fixed-gbp-per-month"])
            else:
                hh["duos-standing-gbp"] = (
                    float(tariff["fixed-gbp-per-day"]) * hh["utc-day"]
                )
        except KeyError as e:
            raise BadRequest(str(e))


def year_md_095(data_source, finish):
    if data_source.site is None:
        return year_md_095_supply(data_source, finish)
    else:
        return year_md_095_site(data_source, finish)


def year_md_095_supply(ds, finish):
    supply = ds.supply
    sess = ds.sess
    md_kva = 0
    month_finish = finish - relativedelta(months=11)

    while not month_finish > finish:
        month_start = month_finish - relativedelta(months=1) + HH
        month_kwh_result = (
            sess.query(func.sum(HhDatum.value), func.max(HhDatum.value))
            .join(Channel)
            .join(Era)
            .filter(
                Era.supply == supply,
                HhDatum.start_date >= month_start,
                HhDatum.start_date <= month_finish,
                Channel.channel_type == "ACTIVE",
                Channel.imp_related == true(),
            )
            .one()
        )

        if month_kwh_result[0] is not None:
            month_md_kw = float(month_kwh_result[1]) * 2
            month_kwh = float(month_kwh_result[0])
            month_kvarh = (
                sess.query(func.sum(HhDatum.value))
                .join(Channel)
                .join(Era)
                .filter(
                    Era.supply == supply,
                    HhDatum.start_date >= month_start,
                    HhDatum.start_date <= month_finish,
                    Channel.channel_type == "REACTIVE_IMP",
                    Channel.imp_related == true(),
                )
                .one()[0]
            )
            if month_kvarh is None:
                pf = 0.95
            else:
                month_kvarh = float(month_kvarh)
                if month_kwh == 0 and month_kvarh == 0:
                    pf = 1
                else:
                    pf = month_kwh / (month_kwh ** 2 + month_kvarh ** 2) ** 0.5
            month_kva = month_md_kw / pf
            md_kva = max(md_kva, month_kva)
        month_finish += relativedelta(months=1)
    return md_kva


def year_md_095_site(data_source, finish, pw):
    md_kva = 0
    month_finish = finish - relativedelta(months=12)

    while not month_finish > finish:
        month_start = month_finish - relativedelta(months=1) + HH
        month_data = {"start": month_start, "finish": month_finish}
        data_source.sum_md(month_data, pw)
        if month_data["sum-kwh"] is None:
            month_md_kw = 0
            month_kwh = 0
        else:
            month_md_kw = month_data["md-kw"]
            month_kwh = month_data["sum-kwh"]

        data_source.sum_md(month_data, pw, False)
        if month_data["sum-kvarh"] is not None:
            month_kvarh = month_data["sum-kvarh"]

        if month_kvarh == 0:
            month_kva = month_md_kw / 0.95
        else:
            if month_kwh == 0 and month_kvarh == 0:
                pf = 1
            else:
                pf = month_kwh / (month_kwh ** 2 + month_kvarh ** 2) ** 0.5
            month_kva = month_md_kw / pf
        md_kva = max(md_kva, month_kva)
        month_finish += relativedelta(months=1)
    return md_kva, ""


def datum_beginning_14(ds, hh):
    rates = get_file_rates(ds.caches, ds.dno_code, hh["start-date"])
    try:
        tariff = rates["tariffs"][ds.llfc_code]
    except KeyError as e:
        raise BadRequest(str(e))

    if 0 < hh["ct-decimal-hour"] <= 7:
        hh["duos-night-kwh"] = hh["msp-kwh"]
        hh["duos-night-gbp"] = hh["msp-kwh"] * float(tariff["night-gbp-per-kwh"])
    else:
        hh["duos-day-kwh"] = hh["msp-kwh"]
        hh["duos-day-gbp"] = hh["msp-kwh"] * float(tariff["day-gbp-per-kwh"])

    if 0 < hh["ct-decimal-hour"] <= 7:
        slot = "night"
    elif (
        hh["ct-day-of-week"] < 5
        and hh["ct-decimal-hour"] > 15.5
        and hh["ct-decimal-hour"] < 19
        and (hh["ct-month"] > 11 or hh["ct-month"] < 4)
    ):
        slot = "winter-weekday-peak"
    elif (
        hh["ct-day-of-week"] < 5
        and (7 <= hh["ct-decimal-hour"] < 16 or 18 < hh["ct-decimal-hour"] < 20)
        and (hh["ct-month"] > 11 or hh["ct-month"] < 4)
    ):
        slot = "winter-weekday-day"
    else:
        slot = "other"

    hh["laf"] = float(rates["lafs"][ds.voltage_level_code.lower()][slot])
    hh["gsp-kwh"] = hh["msp-kwh"] * hh["laf"]
    hh["gsp-kw"] = hh["gsp-kwh"] * 2

    if hh["utc-decimal-hour"] == 0:
        hh["duos-standing-gbp"] = float(tariff["fixed-gbp-per-day"])

    if hh["ct-is-month-end"]:
        month_to = hh["start-date"]
        month_from = month_to - relativedelta(months=1) + HH
        availability = ds.sc
        reactive_rate = float(tariff["reactive-gbp-per-kvarh"])
        hh["duos-reactive-rate"] = reactive_rate
        imp_msp_kvarh = 0
        msp_kwh = 0
        md_kva = 0
        for dsc in ds.get_data_sources(month_from, month_to):
            for h in dsc.hh_data:
                imp_msp_kvarh += h["imp-msp-kvarh"]
                msp_kwh += h["msp-kwh"]
                md_kva = max(
                    md_kva,
                    (h["msp-kw"] ** 2 + (h["imp-msp-kvar"] + h["exp-msp-kvar"]) ** 2)
                    ** 0.5,
                )
        hh["duos-reactive-gbp"] = max(0, imp_msp_kvarh - msp_kwh / 3) * reactive_rate
        if not ds.is_displaced:
            availability_rate = float(tariff["availability-gbp-per-kva-per-day"])
            hh["duos-availability-rate"] = availability_rate
            billed_avail = max(availability, md_kva)
            hh["duos-availability-gbp"] = availability_rate * billed_avail
            hh["duos-availability-agreed-kva"] = ds.sc
            hh["duos-availability-billed-kva"] = billed_avail


def datum_2010_04_01(ds, hh):
    start_date = hh["start-date"]
    dno_cache = ds.caches["dno"][ds.dno_code]

    if not ds.full_channels and not (hh["msp-kwh"] > 0 and hh["anti-msp-kwh"] == 0):
        imp_msp_kvarh, exp_msp_kvarh = 0, 0
    else:
        imp_msp_kvarh, exp_msp_kvarh = hh["imp-msp-kvarh"], hh["exp-msp-kvarh"]

    try:
        gsp_group_cache = dno_cache[ds.gsp_group_code]
    except KeyError:
        gsp_group_cache = dno_cache[ds.gsp_group_code] = {}

    try:
        tariff = gsp_group_cache["tariffs"][ds.llfc_code][start_date]
    except KeyError:
        try:
            tariff_cache = gsp_group_cache["tariffs"]
        except KeyError:
            tariff_cache = gsp_group_cache["tariffs"] = {}

        try:
            tariffs = tariff_cache[ds.llfc_code]
        except KeyError:
            tariffs = tariff_cache[ds.llfc_code] = {}

        try:
            tariff = tariffs[start_date]
        except KeyError:
            tariff = None
            try:
                tariff_list = get_file_rates(ds.caches, ds.dno_code, start_date)[
                    ds.gsp_group_code
                ]["tariffs"]
            except KeyError as e:
                raise BadRequest(str(e))

            for llfcs_pcs, tf in tariff_list.items():
                key = llfcs_pcs.split("_")
                llfcs = [v.strip() for v in key[0].split(",")]
                if len(key) == 2:
                    pcs = [v.strip() for v in key[1].split(",")]
                else:
                    pcs = None

                if ds.llfc_code in llfcs and (pcs is None or ds.pc_code in pcs):
                    tariff = tf
                    break

            if tariff is None:
                raise BadRequest(
                    f"For the DNO {ds.dno_code} and timestamp {hh_format(start_date)} "
                    f"and GSP group {ds.gsp_group_code}, the LLFC '{ds.llfc_code}' "
                    f"can't be found in the 'tariffs' section."
                )

            tariffs[start_date] = tariff

    try:
        band = gsp_group_cache["bands"][start_date]
    except KeyError:
        try:
            bands_cache = gsp_group_cache["bands"]
        except KeyError:
            bands_cache = gsp_group_cache["bands"] = {}

        try:
            band = bands_cache[start_date]
        except KeyError:
            band = "green"
            ct_hr = hh["ct-decimal-hour"]
            weekend = hh["ct-day-of-week"] > 4
            try:
                slots = get_file_rates(ds.caches, ds.dno_code, start_date)[
                    ds.gsp_group_code
                ]["bands"]
            except KeyError as e:
                raise BadRequest(str(e))

            for slot in slots:
                slot_weekend = slot["weekend"] == 1
                if slot_weekend == weekend and slot["start"] <= ct_hr < slot["finish"]:
                    band = slot["band"]
                    break

            bands_cache[start_date] = band

    try:
        laf = dno_cache["lafs"][ds.llfc_code][start_date]
    except KeyError:
        try:
            laf_cache = dno_cache["lafs"]
        except KeyError:
            laf_cache = dno_cache["lafs"] = {}

        try:
            laf_cache_llfc = laf_cache[ds.llfc_code]
        except KeyError:
            laf_cache_llfc = laf_cache[ds.llfc_code] = {}

        try:
            laf = laf_cache_llfc[start_date]
        except KeyError:
            dno_code = ds.dno_code
            if dno_code == "99":
                laf_cache_llfc[start_date] = 1
            else:

                m_start, m_finish = next(
                    c_months_u(
                        start_year=hh["ct-year"], start_month=hh["ct-month"], months=1
                    )
                )
                for (laf,) in ds.sess.execute(
                    select(Laf)
                    .join(Llfc)
                    .join(Party)
                    .where(
                        Party.dno_code == ds.dno_code,
                        Llfc.code == ds.llfc_code,
                        Laf.timestamp >= m_start,
                        Laf.timestamp <= m_finish,
                    )
                ):
                    laf_cache_llfc[laf.timestamp] = float(laf.value)

            try:
                laf = laf_cache_llfc[start_date]
            except KeyError:
                raise BadRequest(
                    f"Missing LAF for DNO {ds.dno_code}, LLFC {ds.llfc_code} and "
                    f"timestamp {hh_format(start_date)}"
                )

    hh["laf"] = laf
    hh["gsp-kwh"] = laf * hh["msp-kwh"]
    hh["gsp-kw"] = hh["gsp-kwh"] * 2

    kvarh = max(
        max(imp_msp_kvarh, exp_msp_kvarh) - (0.95 ** -2 - 1) ** 0.5 * hh["msp-kwh"], 0
    )

    hh["duos-reactive-kvarh"] = kvarh

    duos_reactive_rate = tariff["gbp-per-kvarh"]
    if duos_reactive_rate is not None:
        duos_reactive_rate = float(duos_reactive_rate)
        if duos_reactive_rate != 0:
            hh["duos-reactive-rate"] = duos_reactive_rate
            hh["duos-reactive-gbp"] = kvarh * duos_reactive_rate

    rate = float(tariff[KEYS[band]["tariff-rate"]])
    hh[KEYS[band]["bill-rate"]] = rate
    hh[KEYS[band]["kwh"]] = hh["msp-kwh"]
    hh[KEYS[band]["gbp"]] = rate * hh["msp-kwh"]

    if hh["ct-decimal-hour"] == 23.5 and not ds.is_displaced:
        hh["duos-fixed-days"] = 1
        rate = float(tariff["gbp-per-mpan-per-day"])
        hh["duos-fixed-rate"] = rate
        hh["duos-fixed-gbp"] = rate

        hh["duos-availability-days"] = 1
        kva = ds.sc
        hh["duos-availability-kva"] = kva
        rate = float(tariff["gbp-per-kva-per-day"])
        hh["duos-availability-rate"] = rate
        hh["duos-availability-gbp"] = rate * kva

    if hh["ct-is-month-end"] and not ds.is_displaced:
        month_to = start_date
        month_from = to_utc(ct_datetime(hh["ct-year"], hh["ct-month"], 1))
        md_kva = 0
        days_in_month = 0
        for dsc in ds.get_data_sources(month_from, month_to):
            for datum in dsc.hh_data:
                md_kva = max(
                    md_kva,
                    (
                        datum["msp-kw"] ** 2
                        + max(datum["imp-msp-kvar"], datum["exp-msp-kvar"]) ** 2
                    )
                    ** 0.5,
                )
                if datum["ct-decimal-hour"] == 0:
                    days_in_month += 1

        excess_kva = max(md_kva - ds.sc, 0)

        if "excess-gbp-per-kva-per-day" in tariff and excess_kva != 0:
            rate = float(tariff["excess-gbp-per-kva-per-day"])
            hh["duos-excess-availability-kva"] = excess_kva
            rate = float(tariff["excess-gbp-per-kva-per-day"])
            hh["duos-excess-availability-rate"] = rate
            hh["duos-excess-availability-days"] = days_in_month
            hh["duos-excess-availability-gbp"] = rate * excess_kva * days_in_month


def datum_2012_02_23(ds, hh):
    start_date = hh["start-date"]
    dno_cache = ds.caches["dno"][ds.dno_code]

    if not ds.full_channels and hh["msp-kwh"] == 0:
        imp_msp_kvarh, exp_msp_kvarh = 0, 0
    else:
        imp_msp_kvarh, exp_msp_kvarh = hh["imp-msp-kvarh"], hh["exp-msp-kvarh"]

    try:
        gsp_group_cache = dno_cache[ds.gsp_group_code]
    except KeyError:
        gsp_group_cache = dno_cache[ds.gsp_group_code] = {}

    try:
        tariff = gsp_group_cache["tariffs"][ds.pc_code][ds.llfc_code][start_date]
    except KeyError:
        try:
            tariffs_cache = gsp_group_cache["tariffs"]
        except KeyError:
            tariffs_cache = gsp_group_cache["tariffs"] = {}

        try:
            pc_cache = tariffs_cache[ds.pc_code]
        except KeyError:
            pc_cache = tariffs_cache[ds.pc_code] = {}

        try:
            tariffs = pc_cache[ds.llfc_code]
        except KeyError:
            tariffs = pc_cache[ds.llfc_code] = {}

        try:
            tariff = tariffs[start_date]
        except KeyError:
            tariff = None
            try:
                tariff_list = get_file_rates(ds.caches, ds.dno_code, start_date)[
                    ds.gsp_group_code
                ]["tariffs"]
            except KeyError as e:
                raise BadRequest(str(e))

            for llfcs_pcs, tf in tariff_list.items():
                key = llfcs_pcs.split("_")
                llfcs = [v.strip() for v in key[0].split(",")]
                if len(key) == 2:
                    pcs = [v.strip() for v in key[1].split(",")]
                else:
                    pcs = None

                if ds.llfc_code in llfcs and (pcs is None or ds.pc_code in pcs):
                    tariff = tf
                    break

            if tariff is None:
                raise BadRequest(
                    f"For the DNO {ds.dno_code} and timestamp {hh_format(start_date)} "
                    f"and GSP group {ds.gsp_group_code}, the LLFC {ds.llfc_code} "
                    f"with PC {ds.pc_code} can't be found in the 'tariffs' section."
                )

            tariffs[start_date] = tariff

    try:
        band = gsp_group_cache["bands"][start_date]
    except KeyError:
        try:
            bands_cache = gsp_group_cache["bands"]
        except KeyError:
            bands_cache = gsp_group_cache["bands"] = {}

        try:
            band = bands_cache[start_date]
        except KeyError:
            band = "green"
            ct_hr = hh["ct-decimal-hour"]
            weekend = hh["ct-day-of-week"] > 4
            try:
                slots = get_file_rates(ds.caches, ds.dno_code, start_date)[
                    ds.gsp_group_code
                ]["bands"]
            except KeyError as e:
                raise BadRequest(str(e))

            for slot in slots:
                slot_weekend = slot["weekend"] == 1
                if slot_weekend == weekend and slot["start"] <= ct_hr < slot["finish"]:
                    band = slot["band"]
                    break

            bands_cache[start_date] = band

    try:
        laf = dno_cache["lafs"][ds.llfc_code][start_date]
    except KeyError:
        try:
            laf_cache = dno_cache["lafs"]
        except KeyError:
            laf_cache = dno_cache["lafs"] = {}

        try:
            laf_cache_llfc = laf_cache[ds.llfc_code]
        except KeyError:
            laf_cache_llfc = laf_cache[ds.llfc_code] = {}

        try:
            laf = laf_cache_llfc[start_date]
        except KeyError:
            dno_code = ds.dno_code
            if dno_code in ("88", "99"):
                laf_cache_llfc[start_date] = 1
            else:

                for (laf,) in ds.sess.execute(
                    select(Laf)
                    .join(Llfc)
                    .join(Party)
                    .where(
                        Party.dno_code == ds.dno_code,
                        Llfc.code == ds.llfc_code,
                        Laf.timestamp >= ds.start_date,
                        Laf.timestamp <= ds.finish_date,
                    )
                ):
                    laf_cache_llfc[laf.timestamp] = float(laf.value)

            try:
                laf = laf_cache_llfc[start_date]
            except KeyError:
                for cand in (hh["hist-start"], ds.forecast_date):
                    laf_obj = ds.sess.execute(
                        select(Laf)
                        .join(Llfc)
                        .join(Party)
                        .where(
                            Party.dno_code == ds.dno_code,
                            Llfc.code == ds.llfc_code,
                            Laf.timestamp == cand,
                        )
                    ).scalar_one_or_none()
                    if laf_obj is not None:
                        laf_cache_llfc[start_date] = float(laf_obj.value)
                        break

                try:
                    laf = laf_cache_llfc[start_date]
                except KeyError:
                    raise BadRequest(
                        f"Missing LAF for DNO {ds.dno_code} and LLFC {ds.llfc_code} "
                        f"and timestamps {hh_format(start_date)}, "
                        f"{hh_format(hh['hist-start'])} and "
                        f"{hh_format(ds.forecast_date)}"
                    )

    hh["laf"] = laf
    hh["gsp-kwh"] = laf * hh["msp-kwh"]
    hh["gsp-kw"] = hh["gsp-kwh"] * 2

    kvarh = max(
        max(imp_msp_kvarh, exp_msp_kvarh) - (0.95 ** -2 - 1) ** 0.5 * hh["msp-kwh"], 0
    )

    hh["duos-reactive-kvarh"] = kvarh

    duos_reactive_rate = tariff["gbp-per-kvarh"]
    if duos_reactive_rate is not None:
        duos_reactive_rate = float(duos_reactive_rate)
        if duos_reactive_rate != 0:
            hh["duos-reactive-rate"] = duos_reactive_rate
            hh["duos-reactive-gbp"] = kvarh * duos_reactive_rate

    rate = float(tariff[KEYS[band]["tariff-rate"]])
    hh[KEYS[band]["bill-rate"]] = rate
    hh[KEYS[band]["kwh"]] = hh["msp-kwh"]
    hh[KEYS[band]["gbp"]] = rate * hh["msp-kwh"]

    if hh["ct-decimal-hour"] == 23.5 and not ds.is_displaced:
        hh["duos-fixed-days"] = 1
        rate = float(tariff["gbp-per-mpan-per-day"])
        hh["duos-fixed-rate"] = rate
        hh["duos-fixed-gbp"] = rate

        hh["duos-availability-days"] = 1
        kva = ds.sc
        hh["duos-availability-kva"] = kva
        rate = float(tariff["gbp-per-kva-per-day"])
        hh["duos-availability-rate"] = rate
        hh["duos-availability-gbp"] = rate * kva

    if hh["ct-is-month-end"] and not ds.is_displaced:
        month_to = start_date
        month_from = to_utc(ct_datetime(hh["ct-year"], hh["ct-month"], 1))
        md_kva = 0
        days_in_month = 0
        for dsc in ds.get_data_sources(month_from, month_to):
            for datum in dsc.hh_data:
                md_kva = max(
                    md_kva,
                    (
                        datum["msp-kw"] ** 2
                        + max(datum["imp-msp-kvar"], datum["exp-msp-kvar"]) ** 2
                    )
                    ** 0.5,
                )
                if datum["ct-decimal-hour"] == 0:
                    days_in_month += 1

        excess_kva = max(md_kva - ds.sc, 0)

        if "excess-gbp-per-kva-per-day" in tariff and excess_kva != 0:
            rate = float(tariff["excess-gbp-per-kva-per-day"])
            hh["duos-excess-availability-kva"] = excess_kva
            rate = float(tariff["excess-gbp-per-kva-per-day"])
            hh["duos-excess-availability-rate"] = rate
            hh["duos-excess-availability-days"] = days_in_month
            hh["duos-excess-availability-gbp"] = rate * excess_kva * days_in_month


CUTOFF_DATE_1 = utc_datetime(2010, 3, 31, 23, 0)
CUTOFF_DATE_2 = utc_datetime(2012, 2, 23, 0, 0)


def duos_vb(ds):
    try:
        data_func_cache = ds.caches["dno"][ds.dno_code]["data_funcs"]
    except KeyError:
        try:
            dno_caches = ds.caches["dno"]
        except KeyError:
            dno_caches = ds.caches["dno"] = {}

        try:
            dno_cache = dno_caches[ds.dno_code]
        except KeyError:
            dno_cache = dno_caches[ds.dno_code] = {}

        try:
            data_func_cache = dno_cache["data_funcs"]
        except KeyError:
            data_func_cache = dno_cache["data_funcs"] = {}

    for hh in ds.hh_data:
        try:
            data_func_cache[hh["start-date"]](ds, hh)
        except KeyError:
            if hh["start-date"] < CUTOFF_DATE_1:
                if ds.dno_code == "14":
                    data_func = datum_beginning_14
                elif ds.dno_code == "20":
                    data_func = datum_beginning_20
                elif ds.dno_code in ("22", "99"):
                    data_func = datum_beginning_22
                else:
                    raise Exception("Not recognized")
            elif hh["start-date"] < CUTOFF_DATE_2:
                data_func = datum_2010_04_01
            else:
                data_func = datum_2012_02_23

            data_func_cache[hh["start-date"]] = data_func
            data_func(ds, hh)
